
%    ---------------------------------------------------------------------
%
%    MATLAB/MEX Transforming image completion
%    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%    ---------------------------------------------------------------------

function e = transformic_energy(T,S,Smask,transformation,exact,optimize,patch_w,SsegMask,SignoreMask,Sweight,debug)
% e = transformic_energy(S, Smask, transformation, levels, voting, exact, optimize, SsegMask, SignoreMask, Sweight, name)
% Calculates energy of solution T given source S.
% T: completed (target) image
% S: source image
% Smask: 0, or 1 for pixels to be filled
% transformation: 0 for translation, 1 rotation, 2 similarity
% exact: 0 for patchmatch, 1 to try all locations (optional, default 0)
% optimize parameters: 0 for no, 1 for yes (optional, default 1)
% patch_w: width and height of the square patches (optional, default 7 pixels)
% SsegMask: 0, or i>0 for segments (optional, default all 0s)
% SignoreMask: 0, or 1 for pixels to be ignored (do not fill, do not take patches from) (optional, default all 0s)
% Sweight: weighting to apply to patches around certain pixels (optional, default all 1s)
% debug: set to 1 for debug mode
%
% Alex Mansfield, January 2011
    
    % global variables for optimization
    global opt Sinterp;
    opt = optimset('lsqnonlin'); 
    opt = optimset(opt, 'MaxFunEvals', Inf, 'MaxIter', Inf, 'Display', 'off', 'FunValCheck', 'off');
    opt = optimset(opt, 'Jacobian', 'on'); %, 'DerivativeCheck', 'on');     
    
    % ensure that inputs are double and normalized
    T = double(T);
    S = double(S);
    Smask = double(Smask);
    Smask = Smask(:,:,1);
    if( max(max(Smask))>0 )
        Smask = Smask / max(max(Smask));
    end
    
    if( (nargin<4) || (isempty(transformation)) )
        transformation = 0;
    end
    if( (nargin<5) || (isempty(exact)) )
        exact = 0; % 0 for patchmatch, 1 for exact
    end
    if( (nargin<6) || (isempty(optimize)) )
        optimize = 1; % 0 for no, 1 for yes
    end
    if( (nargin<7) || (isempty(patch_w)) )
        patch_w = 7; % default 7
    end
    if( (nargin<8) || (isempty(SsegMask)) )
        SsegMask = zeros(size(Smask));
    else
        SsegMask = double(SsegMask);
        SsegMask = SsegMask(:,:,1);
    end
    if( (nargin<9) || (isempty(SignoreMask)) )
        SignoreMask = zeros(size(Smask));
    else
        SignoreMask = double(SignoreMask);
        SignoreMask = SignoreMask(:,:,1);
        SignoreMask = SignoreMask / max(max(SignoreMask));
        
        % check that SignoreMask does not override Smask
        SignoreMask(Smask>0)=0;
    end
    if( (nargin<10) || (isempty(Sweight)) )
        Sweight = ones(size(Smask));
    else
        Sweight = double(Sweight);
        Sweight = Sweight(:,:,1);
    end
    if( (nargin<11) || (isempty(debug)) )
        debug = 0;
    end
    
    patch_offset = (patch_w-1)/2;

    [Sgrad{1,1} Sgrad{1,2}] = gradient(S);
    
    S(repmat(Smask|SignoreMask,[1 1 3])==1) = 0;
        
    % initialize output
    Tdomain = imdilate(Smask,ones(patch_w));
  
    % make sure don't take pixels where SignoreMaskh==1, as well as not filling this region
    Smask(SignoreMask>0)=1;
    Tmask = zeros(size(Smask));
    
    % concatenation of S images for interpolation
    Sinterp = double(zeros(size(S,1),size(S,2),9));
    Sinterp(:,:,[1:3]) = S;
    Sinterp(:,:,[4:6]) = Sgrad{1,1};
    Sinterp(:,:,[7:9]) = Sgrad{1,2};
       
    nnfield = transformic_pmnn_oo(S,Sgrad{1,1},Sgrad{1,2},T,patch_w,Smask,Tmask,SsegMask,SsegMask,Sweight,Sweight,Tdomain,transformation,exact,optimize,0,[],0,[],[],[],[],[],[],[],[]);  
    
    e = sum(sum(nnfield(:,:,end))) / sum(sum(Tdomain));
    
end
