
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */

/*  transformic_get_patch.cpp
 *
 *  MEX function for retrieving a single patch
 *
 *  Alex Mansfield 2010, alex.mansfield@vision.ee.ethz.ch
 */ 

#ifdef _WIN32
// these lines added to prevent conflict in definitions of char16_t
#include <yvals.h> 
#if (_MSC_VER >= 1600)
#define __STDC_UTF_16__
#endif
#endif

#include "mex.h"
#include <stdexcept>
#include <string>
#include <cmath>

#include "transformic_pmnn.h"
#include "transformic_get_patch.h"

void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[]) 
{

    char strUsage[] = "Usage: \n\t Spatch = transformic_get_patch(S,nnparams,patch_w,transformation,patchdb,patchind,Sdb,Sindimage)\n\
	Return patch in S specified by nnparams.\n\
	\t\t S:  source image\n\
	\t\t nnparams:  nearest neighbour field at a particular tr,tc\n\
	\t\t patch_w:  patch size\n\
    \t\t transformation: 0 for translation, 1 for rotation, 2 for similarity\n";

    try
    {
	// check number of input/output arguments
    if(nrhs != 8)
	    throw("Needs 8 input arguments S, nnparams, patch_w, transformation, patchdb, patchind, Sdb and Sindimage\n");
	if(nlhs > 1)
	    throw("Takes at most 1 output argument Spatch\n");

	// check number of dimensions
	const int dimsS = mxGetNumberOfDimensions(prhs[0]);
	if(dimsS != 3)
	    throw("S should have 3 dimensions\n");
	const int dimsNN = mxGetNumberOfDimensions(prhs[1]);
	if(dimsNN != 2)
	    throw("nnfield should have no more than 2 dimensions\n");
	const int dimsPatch = mxGetNumberOfDimensions(prhs[2]);
	if(dimsPatch != 2)
	    throw("patch_w should have no more than 2 dimensions\n");
	int dimsTrans = 0;
	if( nrhs>=4 )
	{
	    dimsTrans = mxGetNumberOfDimensions(prhs[3]);
	    if(dimsTrans > 2)
		throw("transformation should have no more than 2 dimensions\n");
	}

	// check dimensions
	const int * sizeS = mxGetDimensions(prhs[0]);
	const int * sizeNN = mxGetDimensions(prhs[1]);
	const int * sizePatch = mxGetDimensions(prhs[2]);
	if((sizePatch[0]!=1)||(sizePatch[1]!=1))
	    throw("patch_w should be scalar\n");
	mwSize const * sizeTrans = NULL;
	if( nrhs>=4 )
	{
	    sizeTrans = mxGetDimensions(prhs[3]);
	    if((sizeTrans[0]!=1)||(sizeTrans[1]!=1))
            throw("transformation should be scalar\n");
	}
	mwSize const * const sizeUse_patchdb = mxGetDimensions(prhs[4]);
	//if((sizeUse_patchdb[0]!=1)||(sizeUse_patchdb[1]!=1))
	//    throw("patchdb should be scalar\n");
	mwSize const * const sizePatchind = mxGetDimensions(prhs[5]);
	//if(sizePatchind[0]!=2)
	//    throw("patchind should be 2 x pN");
	mwSize const * const sizeSdb = mxGetDimensions(prhs[6]);
	//if(sizeSdb[1]!=sizeSind[1])
	//    throw("Sdb should be 3*patchdbS_w^2 x SN");
	mwSize const * const sizeSindimage = mxGetDimensions(prhs[7]);

 	images x;

	// get values
	x.SR = sizeS[0];
	x.SC = sizeS[1];
	x.TR = x.SR;
	x.TC = x.SC;
 	x.patch_w = (int)(*mxGetPr(prhs[2]));
	x.patch_offset = (int)(((float)(x.patch_w-1)) / 2.0);
	x.patchdbT_w = x.patch_w;
	x.patchdbT_offset = x.patch_offset;
	x.patchdbS_w = (int)sqrt((double)sizeSdb[0]/3.0);
	x.patchdbS_offset = (int)(((float)(x.patchdbS_w-1)) / 2.0);
	x.pN = (int)sizePatchind[1];

 	int patch_pixels = x.patch_w*x.patch_w;

	// get input pointers
	x.S = mxGetPr(prhs[0]);
	double * nnparams = mxGetPr(prhs[1]);
	int transformation = (int)(*mxGetPr(prhs[3]));
	int const use_patchdb = (int)(*mxGetPr(prhs[4]));

	x.patchind = NULL;
	x.Sdb = NULL;
	x.Sindimage = NULL;
	if( use_patchdb>0 )
	{
	    x.patchind = mxGetPr(prhs[5]);
	    x.Sdb = mxGetPr(prhs[6]);
	    x.Sindimage = mxGetPr(prhs[7]);
	    patch_pixels = x.pN;
	}

	x.rMat = new double[patch_pixels];
	x.cMat = new double[patch_pixels];

 	x.patchS = new double[patch_pixels*3];
	if(x.patchS == NULL)
	    throw("Could not allocate x.patchS\n");

	// define output arrays
	mwSize sizeOut[2];
	sizeOut[0] = patch_pixels;
	sizeOut[1] = sizeS[2];

	plhs[0] = mxCreateNumericArray(2,sizeOut,mxDOUBLE_CLASS,mxREAL); //initializes to 0
	if(plhs[0] == NULL)
	    throw("Could not allocate output T\n");
	double * Spatchout = mxGetPr(plhs[0]);
    
 	double * Spatch = new double [patch_pixels*sizeS[2]];
	if( transformation==1 )
	    transformic_get_patch<patch_rot_db>(x,nnparams,Spatch);
	else if( transformation==2 )
	    transformic_get_patch<patch_sc_db>(x,nnparams,Spatch);
	else if( transformation==3 )
	    transformic_get_patch<patch_sim_db>(x,nnparams,Spatch);
	else if( transformation==4 )
	    transformic_get_patch<patch_fl_db>(x,nnparams,Spatch);
	else if( transformation==5 )
	    transformic_get_patch<patch_il_db>(x,nnparams,Spatch);
	else if( transformation==6 )
	    transformic_get_patch<patch_simb_db>(x,nnparams,Spatch);
	else 
	    transformic_get_patch<patch_tr_db>(x,nnparams,Spatch); // transformation 0

	double * ptr_Spatch = Spatch;
	for( int i=0; i<patch_pixels; i++ )
	{
		for( int j=0; j<sizeS[2]; j++ )
		{
			Spatchout[index2(i,j,patch_pixels)] = *ptr_Spatch;
			ptr_Spatch++;
		}
	}

	delete [] x.rMat;
	delete [] x.cMat;
	delete [] x.patchS;
	delete [] Spatch;

	return;

    } 
    catch(std::bad_alloc&)
    {
	mexErrMsgTxt("Out of memory");
    }
    catch(const char* p)
    {
	std::string strOut(p);
	strOut += "\n";
	strOut += strUsage;
	mexErrMsgTxt(strOut.c_str());
    }
    catch(...)
    {
	mexErrMsgTxt("Unknown exception caught");
    }

}
