
% T = transformic_inpaint(S, Smask, transformation, levels, voting, 
%                         weightingscheme, exact, optimize, patch_w,
%                         SignoreMask, Sweight, T0, debug,
%                         no_iterations, name, energyfile) 
%
% Fills hole in image S specified by Smask and returns image T
% and nearest neighbour field nnfield.
%
% inputs:
% S:               source image 
% Smask:           binary image with 1 for pixels to be filled, else 0
% transformation:  0 for translation, 1 rotation, 2 scaling, 
%                  3 rotation+scaling, 4 vertical flipping, 5 illumination,
%                  6 all [default 0]
% levels:          number of levels in hierarchy [default 1]
% voting:          pixel colour voting scheme 0 = mean voting, 
%                  1 = median voting, 2 = max voting, 3 = weighted median, 
%                  4 = mode (mean-shift) [default 1]
% weightingscheme: 0 as described in paper, 1 for Wexler paper, 
%                  2 for Wexler paper weighted by evidence [default 0]
% exact:           0 for patchmatch, 1 to try all locations [default 0]
% optimize:        0 for no, 1 for lsqnonlin, 2 with multiple
%                  initializations, 3 gradient descent, 4 our LM,
%                  >4 specifies number of uniformly spaced bins [default 4]
% patch_w:         width and height of the square patches (optional,
%                  default 7 pixels) 
% SignoreMask:     0, or 1 for pixels to be ignored (do not fill,
%                  do not take patches from) (optional, default all 0s)       
% Sweight:         weighting to apply to patches around certain pixels 
%                  (optional but must be >=1, default all 1s) 
% T0:              initial T at lowest resolution 
%                  (optional, default S with black hole) 
% debug:           set to 1 for debug mode 
% no_iterations:   number of patch match iterations [default 10]
% name:            name of image 
%                  (optional, used for saving intermediate output files)   
% energyfile:      path of file to store result energies in
%
% The algorithms this code implements are described in:
% Transforming Image Completion
% Mansfield, Prasad, Rother, Sharp, Kohli and Van Gool
% British Machine Vision Conference 2011
% available at http://www.vision.ee.ethz.ch/~mansfiea/transformic/
% 
% Copyright (C) 2011 Alex Mansfield, alex.mansfield@vision.ee.ethz.ch

%    ---------------------------------------------------------------------
%
%    MATLAB/MEX Transforming image completion
%    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%    ---------------------------------------------------------------------

function [T nnfield] = transformic_inpaint(S,Smask,transformation,levels,voting,weightingscheme,exact,optimize,patch_w,SignoreMask,Sweight,T0,debug,no_iterations,name,energyfile) 
    
% global variables for optimization 
    global opt Sinterp; 
    opt = optimset('lsqnonlin');  
    opt = optimset(opt, 'MaxFunEvals', Inf, 'MaxIter', Inf, 'Display', 'off', 'FunValCheck', 'off'); 
    opt = optimset(opt, 'Jacobian', 'on'); %, 'DerivativeCheck', 'on');      
    
    % ensure that inputs are double and normalized 
    S = double(S); 
    Smask = double(Smask); 
    Smask = Smask(:,:,1); 
    if( max(max(Smask))>0 ) 
        Smask = Smask / max(max(Smask));
    end 
    
    if( (nargin<3) || (isempty(transformation)) ) 
        transformation = 0; 
    end 
    if( (nargin<4) || (isempty(levels)) ) 
        levels = 1; 
    end 
    if( (nargin<5) || (isempty(voting)) ) 
        voting = 1;
    end 
    if( (nargin<6) || (isempty(weightingscheme)) ) 
        weightingscheme = 0;
    end 
    if( (nargin<7) || (isempty(exact)) ) 
        exact = 0;
    end 
    if( (nargin<8) || (isempty(optimize)) ) 
        optimize = 4;
    end 
    if( (nargin<9) || (isempty(patch_w)) ) 
        patch_w = 7;
    end 
    if( (nargin<10) || (isempty(SignoreMask)) ) 
        SignoreMask = zeros(size(Smask)); 
    else 
        SignoreMask = double(SignoreMask); 
        SignoreMask = SignoreMask(:,:,1); 
        SignoreMask = SignoreMask / max(max(SignoreMask)); 
        
        % check that SignoreMask does not override Smask 
        SignoreMask(Smask>0)=0; 
    end 
    if( (nargin<11) || (isempty(Sweight)) ) 
        Sweight = ones(size(Smask)); 
    else 
        Sweight = double(Sweight); 
        Sweight = Sweight(:,:,1); 
    end 
    if( (nargin<12) || (isempty(T0)) ) 
        T0 = []; 
        use_T0 = 0;
    else 
        T0 = double(T0); 
        disp('T0 given'); 
        use_T0 = 1;
    end 
    if( (nargin<13) || (isempty(debug)) ) 
        debug = 0; 
    end 

    % always use patch databases
    patchdb = 1; 

    if( (nargin<13) || (isempty(no_iterations)) ) 
        no_iterations = 10; 
    end
    
    % parameters 
    draw_output = 0; 
    if(debug==1) 
        draw_output = 1; 
    end
    
    save_outputs = 0; 
    if( nargin>14 ) 
        save_outputs = 1; 
    end 
    
    if( (nargin<16) || (isempty(energyfile)) ) 
        energyfile = ''; 
    end 
    
    patch_offset = (patch_w-1)/2; 
    
    % hierarchical 
    Sh = cell(levels,1); 
    Sgradh = cell(levels,2); 
    Th = cell(levels,1); 
    Smaskh = cell(levels,1); 
    Sholeh = cell(levels,1); 
    SignoreMaskh = cell(levels,1); 
    Sweighth = cell(levels,1); 
    
    Sh{1} = S; 
    [Sgradh{1,1} Sgradh{1,2}] = gradient(S); 
    Th{1} = S; 
    
    % set hole and ignored pixels to 0 
    Sh{1}(repmat(Smask|SignoreMask,[1 1 3])==1) = 0; 
    Sgradh{1,1}(repmat(Smask|SignoreMask,[1 1 3])==1) = 0; 
    Sgradh{1,2}(repmat(Smask|SignoreMask,[1 1 3])==1) = 0; 
    Th{1}(repmat(Smask|SignoreMask,[1 1 3])==1) = 0; 
    
    Smaskh{1} = Smask; 
    Sholeh{1} = Smask;
    SignoreMaskh{1} = SignoreMask; 
    Sweighth{1} = Sweight; 
    
    for l = 2:levels 
        Sh{l} = imresize(Sh{1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'bicubic'); 
        Th{l} = imresize(Th{1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'bicubic'); 
        
        % bicubic interpolation may give pixel values out of bounds 
        Sh{l}(Sh{l}<0)=0; 
        Sh{l}(Sh{l}>255)=255; 
        Th{l}(Th{l}<0)=0; 
        Th{l}(Th{l}>255)=255; 
        
        [Sgradh{l,1} Sgradh{l,2}] = gradient(Sh{l}); 
        
        Smaskh{l} = imresize(Smaskh{l-1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'nearest'); 
        Sholeh{l} = imresize(Sholeh{l-1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'nearest'); 
        SignoreMaskh{l} = imresize(SignoreMaskh{l-1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'nearest'); 
        Sweighth{l} = imresize(Sweighth{l-1}, [round(size(Sh{l-1},1)/2) round(size(Sh{l-1},2)/2)], 'bicubic'); 
    end 
    
    for l = 2:levels 
        % set hole and ignored pixels to 0 
        Th{l}(repmat(Smaskh{l}|SignoreMaskh{l},[1 1 3])==1) = 0; 
        Sgradh{l,1}(repmat(Smaskh{l}|SignoreMaskh{l},[1 1 3])==1) = 0; 
        Sgradh{l,2}(repmat(Smaskh{l}|SignoreMaskh{l},[1 1 3])==1) = 0; 
        Sh{l}(repmat(Smaskh{l}|SignoreMaskh{l},[1 1 3])==1) = 0; 
        
        % dilate hole to cover bad incorrect interpolation at hole boundary 
        Smaskh{l} = transformic_imdilate(Smaskh{l},3); 
        Sholeh{l} = transformic_imdilate(Sholeh{l},3); 
    end 
    
    if( isempty(T0) ) 
        T0 = Sh{levels}; 
    end 
    
    % initialize output 
    Th{levels} = T0; 
    Tdomain = transformic_imdilate(Smaskh{levels},patch_w); 
    
    % make sure don't take pixels where SignoreMaskh==1, as well as not filling this region 
    Smaskh{levels}(SignoreMaskh{levels}>0)=1; 
    Smaskh{levels} = double(Smaskh{levels}); 
    Tmask = Smaskh{levels};
    
    if(use_T0)
        Tmask(:) = 0;
    end
    
    % concatenation of S images for interpolation 
    Sinterp = double(zeros(size(Sh{levels},1),size(Sh{levels},2),9)); 
    Sinterp(:,:,[1:3]) = Sh{levels}; 
    Sinterp(:,:,[4:6]) = Sgradh{levels,1}; 
    Sinterp(:,:,[7:9]) = Sgradh{levels,2}; 
    
    % setup patch databases (invariants for all resolutions)
    patchdbT_w = 0;
    patchdbT_offset = 0;
    patchdbS_w = 0;
    patchdbS_offset = 0;
    SN = 0;
    TN = 0;
    pN = 0;
    patchind = zeros(0);
    Sind = zeros(0);
    Tind = zeros(0);
    Sindimage = zeros(0);
    Tindimage = zeros(0);
    Srot = zeros(0);
    Trot = zeros(0);
    rotgaussian = zeros(0);
    Sdb = zeros(0);
    Sgrad_xdb = zeros(0);
    Sgrad_ydb = zeros(0);
    if( patchdb>0 )
        patchdbT_w = patch_w;
        patchdbT_offset = patch_offset;
        
        patchdbS_w = patchdbT_w+2;
        patchdbS_offset = patchdbT_offset+1;
        if( transformation==2 || transformation==3 || (transformation==6) ) %if scaling
            patchdbS_w = patchdbT_w*2-1;
            patchdbS_offset = patchdbT_offset*2;
        end
        
        Sdomain = transformic_imdilate(Smaskh{levels},patchdbS_w);      
        
        if( (transformation==0) || (transformation==4) || (transformation==2) || (transformation==5) )
            % square patches
            [I J] = ind2sub([patch_w patch_w],(1:(patch_w*patch_w)));
            patchind = [I;J] - 1 + patchdbS_offset-patchdbT_offset;
        elseif( transformation==1 || transformation==3 || (transformation==6) )
            % circle patches
            
            % indices into all points 
            [nn,mm] = meshgrid(1:patchdbT_w,1:patchdbT_w);
            % normalize them
            nn = nn - patchdbT_offset - 1;
            mm = mm - patchdbT_offset - 1;

            discmask = ((mm.^2 + nn.^2) <= patchdbT_offset^2);
            
            [I J] = ind2sub([patchdbT_w patchdbT_w],find(discmask>0)');
            patchind = [I;J] - 1 + patchdbS_offset-patchdbT_offset;
        end
        
        pN = length(patchind);
        
        TN = max(0,length(find(Tdomain((patchdbT_offset+1):(end-patchdbT_offset),(patchdbT_offset+1):(end-patchdbT_offset))>0)));
        
        TN_in_SN = length(find(Sdomain((patchdbS_offset+1):(end-patchdbS_offset),(patchdbS_offset+1):(end-patchdbS_offset))>0));
        
        SN = max(0,(size(Sh{levels},1)-2*patchdbS_offset)*(size(Sh{levels},2)-2*patchdbS_offset) - TN_in_SN);
        
        % allocate memory
        Sind = zeros(3,SN);
        Sindimage = -1*ones(size(Smaskh{levels}));
        Tind = zeros(3,TN);
        Tindimage = -1*ones(size(Smaskh{levels}));
        Sdb = zeros(3*patchdbS_w*patchdbS_w,SN);
        if( transformation>0 )
            Sgrad_xdb = zeros(3*patchdbS_w*patchdbS_w,SN);
            Sgrad_ydb = zeros(3*patchdbS_w*patchdbS_w,SN);
        end
        
        % build databases
        Tix = 1;
        for c = (1+patchdbT_offset):(size(Sh{levels},2)-patchdbT_offset)
            for r = (1+patchdbT_offset):(size(Sh{levels},1)-patchdbT_offset)
                if( Tdomain(r,c)~=0 )
                    Tind(:,Tix) = [r-1;c-1;Tix-1];
                    Tindimage(r,c) = Tix-1;
                    Tix = Tix+1;
                end
            end
        end
        
        assert(Tix==(TN+1));
        
        Six = 1;
        for c = (1+patchdbS_offset):(size(Sh{levels},2)-patchdbS_offset)
            for r = (1+patchdbS_offset):(size(Sh{levels},1)-patchdbS_offset)
                if( Sdomain(r,c)==0 )
                    Sind(:,Six) = [r-1;c-1;Six-1];
                    Sindimage(r,c) = Six-1;
                    
                    Spix = 1;
                    for pc = -patchdbS_offset:patchdbS_offset
                        for pr = -patchdbS_offset:patchdbS_offset
                            Sdb(Spix:Spix+2,Six) = Sh{levels}(pr+r,pc+c,:);
                            if( transformation>0 )
                                Sgrad_xdb(Spix:Spix+2,Six) = Sgradh{levels,1}(pr+r,pc+c,:);
                                Sgrad_ydb(Spix:Spix+2,Six) = Sgradh{levels,2}(pr+r,pc+c,:);
                            end
                            Spix = Spix+3;
                        end
                    end
                    Six = Six+1;
                end
            end
        end
        
        assert(Six==(SN+1));
    end
    
    if(draw_output==1) 
        figure(1); image(uint8(S)); 
        title('S'); 
    end 
    
    if(save_outputs>0) 
        imwrite(uint8(Th{levels}),[name '_' int2str(save_outputs) '_i' '.bmp']); 
        save_outputs = save_outputs + 1; 
    end 
    
    % --- Iterate through hierarchy ---------------------------------------------------------------------------------------------
    for l = levels:-1:1 
        convergence_threshold = 0.5; 
        
        % move up tree 
        if(l~=levels) 
            % obtain Tdomain at this level 
            Tdomain = transformic_imdilate(Smaskh{l},patch_w); 
            
            % add in region to ignore to Smaskh 
            Smaskh{l}(SignoreMaskh{l}>0)=1; 
            Smaskh{l} = double(Smaskh{l}); 
            
            % propagate up current Tmask 
            Tmask = imresize(Tmask, [size(Th{l},1) size(Th{l},2)], 'nearest');
            
            % add in region to ignore to Tmask 
            Tmask(SignoreMaskh{l}>0)=1; 
            Tmask = double(Tmask); 
            
            % setup patch databases (at this resolution)
            if( patchdb>0 )
                Sdomain = transformic_imdilate(Smaskh{l},patchdbS_w);
                
                TN = max(0,length(find(Tdomain((patchdbT_offset+1):(end-patchdbT_offset),(patchdbT_offset+1):(end-patchdbT_offset))>0)));
                
                TN_in_SN = length(find(Sdomain((patchdbS_offset+1):(end-patchdbS_offset),(patchdbS_offset+1):(end-patchdbS_offset))>0));

                SN = max(0,(size(Sh{l},1)-2*patchdbS_offset)*(size(Sh{l},2)-2*patchdbS_offset) - TN_in_SN);
                
                % allocate memory
                Sind = zeros(3,SN);
                Sindimage = -1*ones(size(Smaskh{l}));
                Tind = zeros(3,TN);
                Tindimage = -1*ones(size(Smaskh{l}));
                Sdb = zeros(3*patchdbS_w*patchdbS_w,SN);
                if( transformation>0 )
                    Sgrad_xdb = zeros(3*patchdbS_w*patchdbS_w,SN);
                    Sgrad_ydb = zeros(3*patchdbS_w*patchdbS_w,SN);
                end
                
                % build databases
    
                Tix = 1;
                for c = (1+patchdbT_offset):(size(Sh{l},2)-patchdbT_offset)
                    for r = (1+patchdbT_offset):(size(Sh{l},1)-patchdbT_offset)
                        if( Tdomain(r,c)~=0 )
                            Tind(:,Tix) = [r-1;c-1;Tix-1];
                            Tindimage(r,c) = Tix-1;
                            Tix = Tix+1;
                        end
                    end
                end
                
                assert(Tix==(TN+1));
                
                Six = 1;
                for c = (1+patchdbS_offset):(size(Sh{l},2)-patchdbS_offset)
                    for r = (1+patchdbS_offset):(size(Sh{l},1)-patchdbS_offset)
                        if( Sdomain(r,c)==0 )
                            Sind(:,Six) = [r-1;c-1;Six-1];
                            Sindimage(r,c) = Six-1;
                            
                            Spix = 1;
                            for pc = -patchdbS_offset:patchdbS_offset
                                for pr = -patchdbS_offset:patchdbS_offset
                                    Sdb(Spix:Spix+2,Six) = Sh{l}(pr+r,pc+c,:);
                                    if( transformation>0 )
                                        Sgrad_xdb(Spix:Spix+2,Six) = Sgradh{l,1}(pr+r,pc+c,:);
                                        Sgrad_ydb(Spix:Spix+2,Six) = Sgradh{l,2}(pr+r,pc+c,:);
                                    end
                                    Spix = Spix+3;
                                end
                            end
                            Six = Six+1;
                        end
                    end
                end
                
                assert(Six==(SN+1));

            end
            
            % propagate up current nnfield 
            nnfield = imresize(nnfield, [size(Th{l},1) size(Th{l},2)], 'nearest'); 
            nnfield(:,:,[1 2]) = nnfield(:,:,[1 2])*2; % rescale (x,y) parameters 
            
            % propagate up current T using propagated nnfield 
            if( weightingscheme==0 ) 
                Tweight = imresize(Tweight_out, [size(Th{l},1) size(Th{l},2)], 'nearest');  
            elseif( weightingscheme==1 ) 
                Tweight = imresize(Tweight, [size(Th{l},1) size(Th{l},2)], 'nearest');  
            elseif( weightingscheme==2 ) 
                Tweight = imresize(Tweight, [size(Th{l},1) size(Th{l},2)], 'nearest');  
            end             
            % set to 0 where nnfield is not known 
            Tweight([1:(patch_w-1) (end-patch_w+1):end],:) = 0; 
            Tweight(:,[1:(patch_w-1) (end-patch_w+1):end]) = 0; 
            Tweight((patch_w+1):2:(end-patch_w+1),:) = 0; 
            Tweight(:,(patch_w+1):2:(end-patch_w+1)) = 0;             
            
            % vote for new pixel colours 

            if(voting==0) 
            	[Th{l} Tweight_out] = transformic_vote_mean(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==1) 
                [Th{l} Tweight_out] = transformic_vote_median(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==2) 
                [Th{l} Tweight_out] = transformic_vote_max(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==3) 
                [Th{l} Tweight_out] = transformic_vote_weightedmedian(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==4) 
                [Th{l} Tweight_out] = transformic_vote_mode(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            end 

            % set pixels outside holes to correct values 
            NCmask = repmat(Sholeh{l}==0,[1 1 3]); 
            Th{l}(NCmask) = Sh{l}(NCmask); 
            
            % determine pixels still not filled 
            Tmask = double(Tweight_out==0); 
            Tmask(Sholeh{l}==0) = 0;
            
            % resize Sinterp 
            Sinterp = zeros(size(Sh{l},1),size(Sh{l},2),9); 
            
            if(draw_output==1) 
                figure(3); image(uint8(Th{l})); 
                title(['Th\{' int2str(l) '\} after up propagation']); 
                figure(4); image(uint8(Tmask*255)); 
                title(['Tmask']); 
                
                waitforbuttonpress;  
            end 
            
            if(save_outputs>0) 
                imwrite(uint8(Th{l}),[name '_' int2str(save_outputs) '_u' '.bmp']); 
                save_outputs = save_outputs + 1; 
            end 
        end 
        
        if( weightingscheme==0 ) 
            factor = 2; 
        elseif( weightingscheme==1 ) 
            factor = 1.3; 
        elseif( weightingscheme==2 ) 
            factor = 1.3; 
            
            % for amount of evidence conditioned on term 
            SmaskInt = Smaskh{l}; 
            c=1; 
            for r=2:size(SmaskInt,1) 
                SmaskInt(r,c) = SmaskInt(r,c) + SmaskInt(r-1,c);  
            end 
            r=1; 
            for c=2:size(SmaskInt,2) 
                SmaskInt(r,c) = SmaskInt(r,c) + SmaskInt(r,c-1);  
            end 
            for c=2:size(SmaskInt,2) 
                for r=2:size(SmaskInt,1) 
                    SmaskInt(r,c) = SmaskInt(r,c) + SmaskInt(r,c-1) + SmaskInt(r-1,c) - SmaskInt(r-1,c-1);  
                end 
            end 
            
            evidence = zeros(size(SmaskInt)); 
            for c=(patch_offset+1):(size(evidence,2)-patch_offset) 
                for r=(patch_offset+1):(size(evidence,1)-patch_offset) 
                    evidence(r,c) = 1 + (patch_w*patch_w) - SmaskInt(r+patch_offset,c+patch_offset); 
                    if(r>(patch_offset+1)) 
                        evidence(r,c) = evidence(r,c) + SmaskInt(r-patch_offset-1,c+patch_offset); 
                    end 
                    if(c>(patch_offset+1)) 
                        evidence(r,c) = evidence(r,c) + SmaskInt(r+patch_offset,c-patch_offset-1); 
                    end 
                    if((r>(patch_offset+1))&&(c>(patch_offset+1))) 
                        evidence(r,c) = evidence(r,c) - SmaskInt(r-patch_offset-1,c-patch_offset-1); 
                    end 
                end 
            end 
            
        end 
        
        % --- Iterate at current level of the hierarchy -----------------------------------------------------------------------------
        for i = 1:200 
            
            if(i==1 && save_outputs>0)                
                if( exist(name,'file')>0 )
                    load(name);
                end
            else
                if(save_outputs>0) 
                    save(name);
                end 
            end
            
            % copy current output 
            Tprev = Th{l}; 
            
            % concatenation of S images for interpolation 
            Sinterp(:,:,[1:3]) = Sh{l}; 
            Sinterp(:,:,[4:6]) = Sgradh{l,1}; 
            Sinterp(:,:,[7:9]) = Sgradh{l,2}; 
            
            % find nnfield 
            if(l==levels && i==1)
                nnfield = transformic_pmnn_oo(Sh{l},Sgradh{l,1},Sgradh{l,2},Th{l},patch_w,Tmask,Sweighth{l},Sweighth{l},Tdomain,transformation,exact,optimize,0,[],patchdb,patchind,Sind,Tind,Sdb,Sindimage,Tindimage,Sgrad_xdb,Sgrad_ydb,no_iterations);
            else
                nnfield = transformic_pmnn_oo(Sh{l},Sgradh{l,1},Sgradh{l,2},Th{l},patch_w,Tmask,Sweighth{l},Sweighth{l},Tdomain,transformation,exact,optimize,1,nnfield,patchdb,patchind,Sind,Tind,Sdb,Sindimage,Tindimage,Sgrad_xdb,Sgrad_ydb,no_iterations);
            end
            
            if(debug>0)
                keyboard;
            end
            
            if(use_T0==1 && l==1 && i==1)
                energyGT = sum(sum(nnfield(:,:,end)))/sum(sum(Tdomain));
            end
            
            % obtain patch match weighting 
            
            if( weightingscheme==0 ) 
                Tweight = 1-Smaskh{l}; 
                Tdist = bwdist(Tweight); 
                Tweight(Smaskh{l}==1) = factor.^-(double(Tdist(Smaskh{l}==1))); 
                Tweight(Tmask==1) = 0; 
                factor = factor^0.95; 
            elseif( weightingscheme==1 ) 
                % distance transform term 
                Tweight = 1-Smaskh{l}; 
                Tdist = bwdist(Tweight); 
                Tweight(Smaskh{l}==1) = factor.^-(double(Tdist(Smaskh{l}==1))); 
                
                % matching distance term 
                TSdist = nnfield(:,:,end); 

                sigma_ws = prctile(TSdist(TSdist>0),75); 
                Tweight(Smaskh{l}==1) = Tweight(Smaskh{l}==1) .* exp(-TSdist(Smaskh{l}==1).^2 / (2*sigma_ws^2)); 
            elseif( weightingscheme==2 ) 
                % distance transform term 
                Tweight = 1-Smaskh{l}; 
                Tdist = bwdist(Tweight); 
                Tweight(Smaskh{l}==1) = factor.^-(double(Tdist(Smaskh{l}==1))); 
                
                % matching distance term 
                TSdist = nnfield(:,:,end); 
                sigma_ws = prctile(TSdist(TSdist>0),75); 
                Tweight(Smaskh{l}==1) = Tweight(Smaskh{l}==1) .* exp(-TSdist(Smaskh{l}==1).^2 / (2*sigma_ws^2)); 
                
                % amount of evidence conditioned on term 
                Tweight = Tweight.*evidence; 
            end 
            
            % vote for new pixel colours 
            if(voting==0) 
            	[Th{l} Tweight_out] = transformic_vote_mean(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==1) 
                [Th{l} Tweight_out] = transformic_vote_median(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==2) 
                [Th{l} Tweight_out] = transformic_vote_max(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage);    
            elseif(voting==3) 
                [Th{l} Tweight_out] = transformic_vote_weightedmedian(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            elseif(voting==4) 
                [Th{l} Tweight_out] = transformic_vote_mode(Sh{l},nnfield,patch_w,Tweight,Tdomain,transformation,patchdb,patchind,Sdb,Sindimage); 
            end 
            
            % set pixels outside holes to correct values 
            NCmask = repmat(Sholeh{l}==0,[1 1 3]); 
            Th{l}(NCmask) = Sh{l}(NCmask); 
            
            % determine pixels still not filled 
            Tmask = double(Tweight_out==0);
            Tmask(Sholeh{l}==0) = 0;
            
            % set pixels not yet filled inside hole to correct values 
            if( l==levels ) 
                NCmask = repmat(logical(Tmask),[1 1 3]); 
                Th{l}(NCmask) = T0(NCmask); 
                clear NCmask; 
            end 
            
            if(save_outputs>0)      
                imwrite(uint8(Th{l}),[name '_' int2str(save_outputs) '.bmp']); 
                save_outputs = save_outputs + 1; 
            end 
            
            if(draw_output==1) 
                figure(2); image(uint8(Th{l})); 
                title(['Th\{' int2str(l) '\} after voting, (' int2str(i) ')']); 
                
                waitforbuttonpress;
            end 
            
            %check for convergence 
            if(sum(sum(sum((Tprev-Th{l}).^2)))/(size(Th{l},1)*size(Th{l},2)) < convergence_threshold && any(any(Tmask))==0) 
                break;
            end 
            
        end 
        
        if(draw_output==1) 
            figure(2); image(uint8(Th{l})); 
            title(['Th\{' int2str(l) '\} after voting']); 
            
            waitforbuttonpress;  
        end 
    end 
    
    % display energy
    nnfield = transformic_pmnn_oo(Sh{1},Sgradh{1,1},Sgradh{1,2},Th{1},patch_w,Tmask,Sweighth{1},Sweighth{1},Tdomain,transformation,exact,optimize,1,nnfield,patchdb,patchind,Sind,Tind,Sdb,Sindimage,Tindimage,Sgrad_xdb,Sgrad_ydb,no_iterations);
    disp(['Final energy: ' num2str(sum(sum(nnfield(:,:,end)))/sum(sum(Tdomain)))]); 
    T = Th{1};
    
    if(save_outputs>0) 
        imwrite(uint8(T),[name '_end.bmp']); 
  
        [Cnn Ct Cu] = colour_nnfield(nnfield,transformation);
        
        imwrite(Cnn,[name '_cnn.bmp']); 
        if( transformation>0 )
            imwrite(Ct,[name '_ct.bmp']);
        end
        if( transformation==3 )
            imwrite(Cu,[name '_cu.bmp']);
        end
                
        if(use_T0==1)
            fEnergy = fopen(energyfile,'a'); 
            fprintf(fEnergy,'%s_E_%f_T_%f_RMS_%f_EGT_%f\n',name,sum(sum(nnfield(:,:,end)))/sum(sum(Tdomain)),cputime,sqrt(sum(sum(sum((S-T).^2))) / (length(find(Smask(:,:,1)>0))*3)),energyGT); 
            fclose(fEnergy);
        else
            fEnergy = fopen(energyfile,'a'); 
            fprintf(fEnergy,'%s_E_%f_T_%f_RMS_%f\n',name,sum(sum(nnfield(:,:,end)))/sum(sum(Tdomain)),cputime,sqrt(sum(sum(sum((S-T).^2))) / (length(find(Smask(:,:,1)>0))*3))); 
            fclose(fEnergy); 
        end
        
        save([name '_end']);
    end 
    
end 
