
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <math.h>

/**************************************************************************
 * patch_fl
 **************************************************************************/

void patch_fl_db::get_patch(images const & x, double * Simagedb, double * patch) const 
{
    // patch should be allocated length x.pN * 3

    double * ptr_Patchind = x.patchind;
    double * const ptr_Sdb = &(Simagedb[index2(0,s,x.patchdbS_w*x.patchdbS_w*3)]);
    double * ptr_ptr_Sdb;
    int r, c;
    for( int j=0; j<x.pN; j++ )
    {
	r = (int)*ptr_Patchind;
	c = (int)*(++ptr_Patchind);
	ptr_Patchind++;

	if( orient==1 )
	    c = (x.patchdbS_offset-c)+x.patchdbS_offset;

	ptr_ptr_Sdb = &(ptr_Sdb[index3(0,r,c,3,x.patchdbS_w)]);
	for(int k=0; k<3; k++) 
	{
	    *patch = *ptr_ptr_Sdb;
	    patch++;
	    ptr_ptr_Sdb++;
	}
    }

    return;
}

void patch_fl_db::assign_parameters(double * parameters) 
{
    // parameters should be allocated length patch_fl::num_parameters
    sr = (int)parameters[0];
    sc = (int)parameters[1];
    s = (int)parameters[2];
    orient = (int)parameters[3];
}

void patch_fl_db::get_parameters(double * parameters) const 
{
    // parameters should be allocated length patch_fl::num_parameters
    parameters[0] = (double)sr;
    parameters[1] = (double)sc;
    parameters[2] = (double)s;
    parameters[3] = (double)orient;
}

void patch_fl_db::random_parameters(images const & x) 
{
    // random orient
    orient = (int)(rand() % 2);
}

void patch_fl_db::optimize_parameters(images const & x) {
    // optimize orient for current (sr,sc)
    double dist0 = DISTMAX;
    double dist1 = DISTMAX;

    orient = 1;
    calculate_dist(x);
    dist1 = dist;

    orient = 0;
    calculate_dist(x);
    dist0 = dist;

    if( dist1 < dist0 )
	orient = 1;
    
}

void patch_fl_db::random(images const & x, double const alpha) {
    // random

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s, best_orient = orient;
    double best_dist = dist;

    // random search
    double search_radius = (double)std::max(x.TC,std::max(x.TR,std::max(x.SC,x.SR)));
    while( search_radius > 1.0 )
    {
	sr += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));
	sc += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));

	// if this points to a valid patch
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		else
		    random_parameters(x);
		
		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    orient = best_orient;
		    dist = best_dist;
		} else {
		    // if have found a better match, update best
		    best_sr = sr;
		    best_sc = sc;
		    best_s = s;
		    best_orient = orient;
		    best_dist = dist;
		}	
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		orient = best_orient;
		dist = best_dist;
	    }						
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    orient = best_orient;
	    dist = best_dist;
	}

	search_radius *= alpha;
    }
}

void patch_fl_db::consider(patch_fl_db const & p, images const & x) {
   // consider parameters based on parameters at another location

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s, best_orient = orient;
    double best_dist = dist;

    int p_location[3];
    double p_parameters[patch_fl_db::num_parameters+1];
    p.get_location(p_location);
    p.get_parameters(p_parameters);

    if( p_location != NULL && p_parameters != NULL )
    {
	// p_location and p_parameters assigned
	orient = (int)p_parameters[3];
	sr = (int)(tr - p_location[0]) + (int)p_parameters[0];
	sc = (int)(tc - p_location[1]) + (int)p_parameters[1];
	if( orient==1 )
	    sc = (int)(p_location[1] - tc) + (int)p_parameters[1];

	// check transformed patch match from S is valid
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    orient = best_orient;
		    dist = best_dist;
		}
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		orient = best_orient;
		dist = best_dist;	
	    }
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    orient = best_orient;
	    dist = best_dist;
	}
    } //else do nothing
}
