
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <math.h>

#include "mex.h"

#define PI 3.14159265

/**************************************************************************
 * patch_rot
 **************************************************************************/

void patch_rot_db::get_patch(images const & x, double * Simagedb, double * patch) const 
{
    // patch should be allocated length x.pN * 3

    // apply rotation and offset to patch indices

    // theta is in range [0,8]
    double ct = cos((theta / 4.0) * PI);
    double st = sin((theta / 4.0) * PI);

    double * ptr_Patchind = x.patchind;
    double * ptr_rMat = x.rMat;
    double * ptr_cMat = x.cMat;

    double r, c;
    for( int j=0; j<x.pN; j++ )
    {
	r = *ptr_Patchind - (double)x.patchdbS_offset;
	c = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	*ptr_rMat = ct*r + st*c + (double)x.patchdbS_offset + 1.0;
	*ptr_cMat = ct*c - st*r + (double)x.patchdbS_offset + 1.0;

	ptr_rMat++;
	ptr_cMat++;
    }

    vgg_interp2_c_db(patch,&(Simagedb[index2(0,s,x.patchdbS_w*x.patchdbS_w*3)]),x.patchdbS_w,x.pN,x.cMat,x.rMat);

    //mexPrintf("(theta %f)",theta);

    return;
}

void patch_rot_db::assign_parameters(double * parameters) 
{
    // parameters should be allocated length patch_rot::num_parameters
    sr = (int)parameters[0];
    sc = (int)parameters[1];
    s = (int)parameters[2];
    theta = parameters[3];
}

void patch_rot_db::get_parameters(double * parameters) const 
{
    // parameters should be allocated length patch_rot::num_parameters
    parameters[0] = (double)sr;
    parameters[1] = (double)sc;
    parameters[2] = (double)s;
    parameters[3] = theta;
}

void patch_rot_db::random_parameters(images const & x) 
{
    if( x.optimize>4 )
    {
	// random discrete theta
	theta = ((double)(rand() % x.optimize)*(8.0/((double)x.optimize)));
    } else {
	// random continous theta
	// theta is in range [0,8]
	theta = 8.0*((double)rand()/(double)RAND_MAX);
    }    
}

int patch_rot_db::gradient(images const & x) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    int retval = 1;

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    // determine rotation gradient matrix
    double ct = cos((theta / 4.0) * PI);
    double st = sin((theta / 4.0) * PI);

    double ind_x, ind_y, ind_x_grad, ind_y_grad, g=0.0, this_g=0.0;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	ind_x_grad = -st*ind_x - ct*ind_y;
	ind_y_grad = ct*ind_x - st*ind_y;
	
	this_g = 0.0;
	for( int k=0; k<3; k++ )
	{
	    this_g += (*ptr_patchSgradx*ind_x_grad + *ptr_patchSgrady*ind_y_grad) * (*ptr_Spatch-*ptr_Tpatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	g += (1.0 - *ptr_Tmask) * this_g;
	
	ptr_Tmask++;
    }

    // return value (sign of sum of gradient components)
    if(g<0)
	retval = -1;
    
    return retval;
}

void patch_rot_db::gradient_lm(images const & x, double * err, double * grad) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    // err and grad should be allocated to length 3*pN

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    // determine rotation gradient matrix
    double ct = cos((theta / 4.0) * PI);
    double st = sin((theta / 4.0) * PI);

    double ind_x, ind_y, ind_x_grad, ind_y_grad;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	ind_x_grad = -st*ind_x - ct*ind_y;
	ind_y_grad = ct*ind_x - st*ind_y;
	
	for( int k=0; k<3; k++ )
	{
	    //grad[index2(k,i,3)] = (1.0-*ptr_Tmask)*((*ptr_patchSgradx*ind_x_grad + *ptr_patchSgrady*ind_y_grad) * (*ptr_Spatch-*ptr_Tpatch));
	    grad[index2(k,i,3)] = (1.0-*ptr_Tmask)*((*ptr_patchSgradx*ind_x_grad + *ptr_patchSgrady*ind_y_grad));
	    err[index2(k,i,3)] =  (1.0-*ptr_Tmask)*(*ptr_Tpatch - *ptr_Spatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	ptr_Tmask++;
    }
}

void patch_rot_db::optimize_parameters(images const & x) {
//     optimize theta for current (sr,sc)

    if( x.optimize==1 )
    {
	// populate mxArrays
	for( int k=0; k<3; k++ )
	{
	    for( int c=-x.patch_offset; c<=x.patch_offset; c++ )
	    {
		for (int r=-x.patch_offset; r<=x.patch_offset; r++ )
		{
		    x.ptr_Tpatch[index2((r+x.patch_offset) + (c+x.patch_offset)*x.patch_w + k*(x.patch_w*x.patch_w),0,3*x.patch_w*x.patch_w)] = x.T[index3(r+tr,c+tc,k,x.TR,x.TC)];

		    // note that need the inverse mask (0 for pixels to ignore)
		    x.ptr_Tmask[index2((r+x.patch_offset) + (c+x.patch_offset)*x.patch_w + k*(x.patch_w*x.patch_w),0,3*x.patch_w*x.patch_w)] = 1-x.Tmask[index2(r+tr,c+tc,x.TR)];
		}
	    }
	}

	// construct mxArrays
	mxArray * mxsr, * mxsc, * mxtheta;
	mxsr = mxCreateDoubleScalar((double)sr);
	mxsc = mxCreateDoubleScalar((double)sc);
	mxtheta = mxCreateDoubleScalar((theta / 4.0) * PI);

	// call lsqnonlin through transformic_patchmatch_rot
	mxArray * rhsLSQ[6];
	
	rhsLSQ[0] = x.mxTpatch;
	rhsLSQ[1] = x.mxTmask;
	rhsLSQ[2] = x.mxIxpatch;
	rhsLSQ[3] = mxsr;
	rhsLSQ[4] = mxsc;
	rhsLSQ[5] = mxtheta;

	mxArray * lhsLSQ[1];
	if( mexCallMATLAB(1, lhsLSQ, 6, rhsLSQ,"transformic_patchmatch_rot")!=0 )
	{
	    // handle error of call not working
	}
	theta = ((*mxGetPr(lhsLSQ[0])) / PI) * 4.0;
    } 
    else if( x.optimize==2 )
    {
	int no_theta = 6;

	// populate mxArrays
	for( int k=0; k<3; k++ )
	{
	    for( int c=-x.patch_offset; c<=x.patch_offset; c++ )
	    {
		for (int r=-x.patch_offset; r<=x.patch_offset; r++ )
		{
		    x.ptr_Tpatch[index2((r+x.patch_offset) + (c+x.patch_offset)*x.patch_w + k*(x.patch_w*x.patch_w),0,3*x.patch_w*x.patch_w)] = x.T[index3(r+tr,c+tc,k,x.TR,x.TC)];

		    // note that need the inverse mask (0 for pixels to ignore)
		    x.ptr_Tmask[index2((r+x.patch_offset) + (c+x.patch_offset)*x.patch_w + k*(x.patch_w*x.patch_w),0,3*x.patch_w*x.patch_w)] = 1-x.Tmask[index2(r+tr,c+tc,x.TR)];
		}
	    }
	}

	// construct mxArrays
	mxArray * mxsr, * mxsc, * mxtheta;
	mxsr = mxCreateDoubleScalar((double)sr);
	mxsc = mxCreateDoubleScalar((double)sc);
	mxtheta = mxCreateDoubleScalar((theta / 4.0) * PI);

	// call lsqnonlin through transformic_patchmatch_rot
	mxArray * rhsLSQ[6];
	
	rhsLSQ[0] = x.mxTpatch;
	rhsLSQ[1] = x.mxTmask;
	rhsLSQ[2] = x.mxIxpatch;
	rhsLSQ[3] = mxsr;
	rhsLSQ[4] = mxsc;
	rhsLSQ[5] = mxtheta;

	mxArray * lhsLSQ[2];
	double curr_theta = theta;
	double best_theta = theta;
	double curr_dist = DISTMAX;
	double best_dist = DISTMAX;
	for(int i=1; i<=no_theta; i++)
	{
	    if( mexCallMATLAB(2, lhsLSQ, 6, rhsLSQ,"transformic_patchmatch_rot")!=0 )
	    {
		// handle error of call not working
	    }
	    curr_theta = ((*mxGetPr(lhsLSQ[0])) / PI) * 4.0;
	    curr_dist = *mxGetPr(lhsLSQ[1]);

	    // found a better theta?
	    if( curr_dist < best_dist)
	    {
		best_theta = curr_theta;
		best_dist = curr_dist;
	    }

	    // new intiial theta
	    curr_theta = theta + i*(8.0/(double)no_theta);
	    if(curr_theta >= 8.0)
		curr_theta -= 8.0;
	    mxtheta = mxCreateDoubleScalar((curr_theta / 4.0) * PI);
	    rhsLSQ[5] = mxtheta;
	}
	
	theta =  best_theta;
    }
    else if( x.optimize==3 )
    {
	// c++ gradient descent algorithm

	double const initial_theta = theta;

	// parameters
	int const no_theta = 6;
	double const precision = PI / 100000.0;
	double const initial_step = PI / (double)no_theta; //this should be positive

	// for storing the best
	double theta_best = initial_theta;
	double dist_best = DISTMAX;
	
	// initialize variables
	double theta_old = initial_theta;
	double dist_old = DISTMAX;
	dist = DISTMAX;
	double step_old = initial_step;
	double step = initial_step;
	int direction = 1;

	// for a range of starting values
	for(int i=0; i<no_theta; i++)
	{
	  // try new initial theta
	  theta_old = initial_theta + ((double)(i)*(8.0/((double)no_theta)));
	  if(theta_old >= 8.0)
	    theta_old -= 8.0;
	  theta = theta_old;
  
	  // initialize variables
	  dist_old = DISTMAX;
	  dist = DISTMAX;
	  step_old = initial_step;
	  step = initial_step;
	  direction = 1;

	  //mexPrintf("initial %f",theta);
	  // while step size still larger than precision
	  while( fabs(step)>precision )
	  {
	      calculate_dist(x);
	      
	      // if cost(theta) < cost(theta_old)
	      if( dist<dist_old )
	      {
		  theta_old = theta;
		  dist_old = dist;
		  step_old = fabs(step);
		  step = step_old;
		  
		  // set direction based on this theta
		  direction = -1*gradient(x);
		  
		  // if best result so far, store
		  if( dist < dist_best )
		  {
		      theta_best = theta;
		      dist_best = dist;
		  }
		  
	      } else {
		  // halve the step size and try again
		  step /= 2.0;
		  
		  // if we have halved the step size to the precision level
		  // try flipping the direction
		  if( (fabs(step)<=precision) && (step>0) )
		      step = -1*step_old;
	      }
	      
	      // new proposal theta
	      theta = theta_old + ((double)direction)*step;
	      
	      // keep within bounds
	      if(theta >= 8.0)
		  theta -= 8.0;
	      if(theta < 0.0)
		  theta += 8.0;
	      
	  }
	  //mexPrintf(" %i %f (%f)",i,theta_old,dist_old);
	  
	}
	
	theta =  theta_best;
	dist = dist_best;

    } else if( x.optimize==4 )
    {
	// Levenburg-Marquadt

	// parameters
	int const no_theta = 6;
	double const eps_1 = 1e-5;
	double const eps_2 = 1e-5;
	double const eps_3 = 1e-5;
	double const tau = 1e-2;
	int const kmax = 100;

	// for storing the best
	double theta_best = theta;
	double dist_best = DISTMAX;

	// for storing bounds
	double theta_max = 8.0;
	double theta_min = 0.0;

	double * err = new double[3*x.pN];
	double * grad = new double[3*x.pN];

	double const initial_theta = theta;
	double diveight = 0.0;

	//TEST CODE********************************
	// if( tc==74 && tr==42 && sc==77 && sr==74 )
// 	{
// 	    int test_no_theta = 120;
// 	    for(int thetas=0; thetas<test_no_theta; thetas++)
// 	    {
// 		// try new initial theta
// 		theta = initial_theta + ((double)(thetas)*(8.0/((double)test_no_theta)));

// 		calculate_dist(x);

// 		mexPrintf("%f,%f;\n",theta,dist);
// 	    }

// 	    //TEST CODE********************************

	for(int thetas=0; thetas<no_theta; thetas++)
	{
	    // try new initial theta
	    theta = initial_theta + ((double)(thetas)*(8.0/((double)no_theta)));
	    
	    // having found theta, set up the bounds
	    theta_max = theta + 4.0/(double)no_theta;
	    theta_min = theta - 4.0/(double)no_theta;

	    double nu = 2.0;

	    calculate_dist(x);

	    // mexPrintf("%f,%f,",theta,dist);

	    gradient_lm(x,err,grad);

	    double A = 0;
	    double g = 0;
	    double * ptr_err = err;
	    double * ptr_grad = grad;
	    for( int i=0; i<x.pN; i++ )
	    {
		for( int j=0; j<3; j++ )
		{
		    g += (*ptr_grad)*(*ptr_err);
		    A += (*ptr_grad)*(*ptr_grad);
		    
		    ptr_err++;
		    ptr_grad++;
		}
	    }

	    int stop = 0;
	
	    if( fabs(g) < eps_1 )
		stop = 1;

	    double mu = tau * A;

	    int k = 0;
	    while( stop==0 && (k<kmax) )
	    {
		k++;

		double rho = 0.0;
		while( stop==0 && rho<=0.0 )
		{
		    double delta_theta = 0.0;
		    if( g != 0.0 )
			delta_theta = g / (A+mu);

		    //if( fabs(delta_theta) <= eps_2*fabs(theta) ) alex remove
		    if( fabs(delta_theta) <= eps_2 ) // alex add
		    {
			stop = 1;
		    } else {
			double theta_old = theta;
			theta += delta_theta;

			//thresholds to bounds and stop (this is the local minimum)
			if( (theta>=theta_max) || (theta<=theta_min) )
			{
			    theta = std::max(std::min(theta,theta_max),theta_min);
			    stop = 1;
			}

			double dist_old = dist; 
			calculate_dist(x);
			
			rho = ((dist_old - dist)/(delta_theta*(mu*delta_theta + g)));
			if( rho>0 )
			{
			    gradient_lm(x,err,grad);

			    A = 0;
			    g = 0;
			    ptr_err = err;
			    ptr_grad = grad;
			    for( int i=0; i<x.pN; i++ )
			    {
				for( int j=0; j<3; j++ )
				{
				    g += (*ptr_grad)*(*ptr_err);
				    A += (*ptr_grad)*(*ptr_grad);
			   
				    ptr_err++;
				    ptr_grad++;
				}
			    }

			    if( (fabs(g) < eps_1) || (dist <= eps_3) )
				stop = 1;

			    mu *= std::max(1.0/3.0,1-(2.0*rho-1.0)*(2.0*rho-1.0)*(2.0*rho-1.0));
			    nu = 2.0;
			
			} else {
			    theta = theta_old;
			    dist = dist_old;

			    mu *= nu;
			    nu *= 2.0;
			}
		    }
		}
	    }

	    if( dist < dist_best )
	    {
		dist_best = dist;
		theta_best = theta;
	    }

	    //mexPrintf("%f,%f;\n",theta,dist);
	}

	dist = dist_best;
	diveight = theta_best/8.0;
	theta_best -= 8.0*(double)floor(diveight);
	theta = theta_best;

	//TEST CODE********************************
	// }
	//TEST CODE********************************

	delete [] err;
	delete [] grad;
    } else if( x.optimize>4 )
    {
	// x.optimize number of bins
	double diveight = 0.0;

	double best_theta = theta, best_dist = dist;

	for(int thetas=0; thetas<x.optimize; thetas++)
	{
	    // try new initial theta
	    theta = ((double)(thetas)*(8.0/((double)x.optimize)));

	    calculate_dist(x);

	    if( dist < best_dist )
	    {
		best_dist = dist;
		best_theta = theta;
	    }
	}

	diveight = best_theta/8.0;
	best_theta -= 8.0*(double)floor(diveight);
	theta = best_theta;
	dist = best_dist;
    }

    // check patch generated in c code is the same
    /*mexPrintf("C code\n");
    double * spatch;
    spatch = new double[x.patch_w * x.patch_w * 3];
    get_patch(x,x.S,3,spatch);
    for( int k=0; k<3; k++ )
    {
	for( int r=-x.patch_offset; r<=x.patch_offset; r++ )
	{
	    for (int c=-x.patch_offset; c<=x.patch_offset; c++ )
	    {
		mexPrintf("%f ",spatch[index3(r+x.patch_offset,c+x.patch_offset,k,x.patch_w,x.patch_w)]);
	    }
	    mexPrintf("\n");
	}
	mexPrintf("\n");
 	}*/

}

void patch_rot_db::random(images const & x, double const alpha) {
    // random

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_theta = theta, best_dist = dist;

    // random search
    double search_radius = (double)std::max(x.TC,std::max(x.TR,std::max(x.SC,x.SR)));
    double search_radius_theta = 4.0;
    while( search_radius > 1.0 )
    {
	sr += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));
	sc += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));

	// if this points to a valid patch
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		else
		{
		    //random_parameters(x);

		    // GENERALIZED PATCH MATCH
		    theta += search_radius_theta*((2.0*((double)rand()/(double)RAND_MAX))-1.0);
		    double diveight = theta/8.0;
		    theta -= 8.0*(double)floor(diveight);
		    
		}

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    theta = best_theta;
		    dist = best_dist;
		} else {
		    // if have found a better match, update best
		    best_sr = sr;
		    best_sc = sc;
		    best_s = s;
		    best_theta = theta;
		    best_dist = dist;
		}	
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		theta = best_theta;
		dist = best_dist;
	    }						
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    theta = best_theta;
	    dist = best_dist;
	}

	search_radius *= alpha;
	search_radius_theta *= (1.5*alpha);
    }
}

void patch_rot_db::consider(patch_rot_db const & p, images const & x) {
   // consider parameters based on parameters at another location

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_theta = theta, best_dist = dist;

    int p_location[3];
    double p_parameters[patch_rot_db::num_parameters+1];
    p.get_location(p_location);
    p.get_parameters(p_parameters);

    if( p_location != NULL && p_parameters != NULL )
    {
	// p_location and p_parameters assigned

	theta = p_parameters[3];
	double ct = cos(theta / 4.0 * PI);
	double st = sin(theta / 4.0 * PI);
	sr = (int)(ct*(tr - p_location[0]) + st*(tc - p_location[1])) + (int)p_parameters[0];
	sc = (int)(ct*(tc - p_location[1]) - st*(tr - p_location[0])) + (int)p_parameters[1];

	// check transformed patch match from S is valid
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    theta = best_theta;
		    dist = best_dist;
		}
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		theta = best_theta;
		dist = best_dist;	
	    }
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    theta = best_theta;
	    dist = best_dist;
	}
    } //else do nothing
}
