
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <math.h>

#include "mex.h"

/**************************************************************************
 * patch_sc
 **************************************************************************/

void patch_sc_db::get_patch(images const & x, double * Simagedb, double * patch) const 
{
    // patch should be allocated length x.pN * 3

    // apply scale and offset to patch indices

    double * ptr_Patchind = x.patchind;
    double * ptr_rMat = x.rMat;
    double * ptr_cMat = x.cMat;

    double r, c;
    for( int j=0; j<x.pN; j++ )
    {
	r = *ptr_Patchind - (double)x.patchdbS_offset;
	c = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	*ptr_rMat = (1.0+scale)*r + (double)x.patchdbS_offset + 1.0;
	*ptr_cMat = (1.0+scale)*c + (double)x.patchdbS_offset + 1.0;

	ptr_rMat++;
	ptr_cMat++;
    }

    vgg_interp2_c_db(patch,&(Simagedb[index2(0,s,x.patchdbS_w*x.patchdbS_w*3)]),x.patchdbS_w,x.pN,x.cMat,x.rMat);

    return;
}

void patch_sc_db::assign_parameters(double * parameters) 
{
    // parameters should be allocated length patch_rot::num_parameters
    sr = (int)parameters[0];
    sc = (int)parameters[1];
    s = (int)parameters[2];
    scale = parameters[3];
}

void patch_sc_db::get_parameters(double * parameters) const 
{
    // parameters should be allocated length patch_rot::num_parameters
    parameters[0] = (double)sr;
    parameters[1] = (double)sc;
    parameters[2] = (double)s;
    parameters[3] = scale;
}

void patch_sc_db::random_parameters(images const & x) 
{
    if( x.optimize>4 )
    {
	// random discrete scale
	// 0<scale<=1.0
	scale = ((double)(rand() % x.optimize))*(1.0/((double)x.optimize));
    } else {
	// random continous scale
	// 0<scale<=1.0
	scale = ((double)rand()+1.0)/((double)RAND_MAX+1.0);
    }
}

int patch_sc_db::gradient(images const & x) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    int retval = 1;

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    double ind_x, ind_y, g=0.0, this_g=0.0;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	this_g = 0.0;
	for( int k=0; k<3; k++ )
	{
	    this_g += (*ptr_patchSgradx*ind_x + *ptr_patchSgrady*ind_y) * (*ptr_Spatch-*ptr_Tpatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	g += (1.0-*ptr_Tmask) * this_g;
	
	ptr_Tmask++;
    }

    // return value (sign of sum of gradient components)
    if(g<0)
	retval = -1;
    
    return retval;
}

void patch_sc_db::gradient_lm(images const & x, double * err, double * grad) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    // err and grad should be allocated to length 3*pN

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    double ind_x, ind_y;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	for( int k=0; k<3; k++ )
	{
	    grad[index2(k,i,3)] = (1.0-*ptr_Tmask)*((*ptr_patchSgradx*ind_x + *ptr_patchSgrady*ind_y));
	    err[index2(k,i,3)] =  (1.0-*ptr_Tmask)*(*ptr_Tpatch - *ptr_Spatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	ptr_Tmask++;
    }
}

void patch_sc_db::optimize_parameters(images const & x) {
//     optimize scale for current (sr,sc)

    if( x.optimize==3 )
    {
	// c++ gradient descent algorithm

	double const initial_scale = scale;

	// parameters
	int const no_scale = 6;
	double const precision = 1.0 / 100000.0;
	double const initial_step = 1.0 / (double)no_scale; //this should be positive

	// for storing the best
	double scale_best = initial_scale;
	double dist_best = DISTMAX;
	
	// initialize variables
	double scale_old = initial_scale;
	double dist_old = DISTMAX;
	dist = DISTMAX;
	double step_old = initial_step;
	double step = initial_step;
	int direction = 1;

	// for a range of starting values
	for(int i=0; i<no_scale; i++)
	{
	    // try new initial theta
	    scale_old = initial_scale + ((double)(i)/((double)no_scale));
	    scale_old = std::max(std::min(scale_old,1.0),0.0);
	    scale = scale_old;
  
	    // initialize variables
	    dist_old = DISTMAX;
	    dist = DISTMAX;
	    step_old = initial_step;
	    step = initial_step;
	    direction = 1;

	    //mexPrintf("initial %f",theta);
	    // while step size still larger than precision
	    while( fabs(step)>precision )
	    {
		calculate_dist(x);
	      
		// if cost(theta) < cost(theta_old)
		if( dist<dist_old )
		{
		    scale_old = scale;
		    dist_old = dist;
		    step_old = fabs(step);
		    step = step_old;
		  
		    // set direction based on this theta
		    direction = -1*gradient(x);
		  
		    // if best result so far, store
		    if( dist < dist_best )
		    {
			scale_best = scale;
			dist_best = dist;
		    }
		  
		} else {
		    // halve the step size and try again
		    step /= 2.0;
		  
		    // if we have halved the step size to the precision level
		    // try flipping the direction
		    if( (fabs(step)<=precision) && (step>0) )
			step = -1*step_old;
		}
	      
		// new proposal theta
		scale = scale_old + ((double)direction)*step;
	      
		// keep within bounds
		scale = std::max(std::min(scale,1.0),0.0);
	      
	    }
	    //mexPrintf(" %i %f (%f)",i,theta_old,dist_old);
	  
	}
	
	scale = scale_best;
	dist = dist_best;

    } else if( x.optimize==4 )
    {
	// Levenburg-Marquadt

	// parameters
	int const no_scale = 6;
	double const eps_1 = 1e-5;
	double const eps_2 = 1e-5;
	double const eps_3 = 1e-5;
	double const tau = 1e-2;
	int const kmax = 100;

	// for storing the best
	double scale_best = scale;
	double dist_best = DISTMAX;

	double * err = new double[3*x.pN];
	double * grad = new double[3*x.pN];

	double const initial_scale = scale;
	for(int scales=0; scales<no_scale; scales++)
	{
	    // try new initial scale
	    scale = initial_scale + ((double)(scales)*(1.0/((double)no_scale)));
	    // to try lower scales than initial scale, modulo 1
	    scale -= (double)floor(scale);
	    
	    double nu = 2.0;

	    calculate_dist(x);
	    gradient_lm(x,err,grad);

	    double A = 0;
	    double g = 0;
	    double * ptr_err = err;
	    double * ptr_grad = grad;
	    for( int i=0; i<x.pN; i++ )
	    {
		for( int j=0; j<3; j++ )
		{
		    g += (*ptr_grad)*(*ptr_err);
		    A += (*ptr_grad)*(*ptr_grad);
		    
		    ptr_err++;
		    ptr_grad++;
		}
	    }

	    int stop = 0;
	
	    if( fabs(g) < eps_1 )
	        stop = 1;

	    double mu = tau * A;

	    int k = 0;
	    while( stop==0 && (k<kmax) )
	    {
		k++;

		double rho = 0.0;
		while( stop==0 && rho<=0.0 )
		{
		    double delta_scale = 0.0;
		    if( g != 0.0 )
			delta_scale = g / (A+mu);

		    if( fabs(delta_scale) <= eps_2 )
		    {
			stop = 1;
		    } else {
			double scale_old = scale;
			scale += delta_scale;
			//thresholds to bounds and stop (this is the local minimum)
			if( (scale>=0.0) || (scale<=1.0) )
			{
			    scale = std::max(std::min(scale,1.0),0.0);
			    stop = 1;
			}

			double dist_old = dist; 
		        calculate_dist(x);
			
			rho = ((dist_old - dist)/(delta_scale*(mu*delta_scale + g)));
			if( rho>0 )
			{
			    gradient_lm(x,err,grad);

			    A = 0;
			    g = 0;
			    ptr_err = err;
			    ptr_grad = grad;
			    for( int i=0; i<x.pN; i++ )
			    {
				for( int j=0; j<3; j++ )
				{
				    g += (*ptr_grad)*(*ptr_err);
				    A += (*ptr_grad)*(*ptr_grad);
			   
				    ptr_err++;
				    ptr_grad++;
				}
			    }

			    if( (fabs(g) < eps_1) || (dist <= eps_3) )
				stop = 1;

			    mu *= std::max(1.0/3.0,1-(2.0*rho-1.0)*(2.0*rho-1.0)*(2.0*rho-1.0));
			    nu = 2.0;
			
			} else {
			    scale = scale_old;
			    dist = dist_old;

			    mu *= nu;
			    nu *= 2.0;
			}
		    }
		}
	    }

	    if( dist < dist_best )
	    {
		dist_best = dist;
		scale_best = scale;
	    }
	}

	dist = dist_best;
	scale = scale_best;

	delete [] err;
	delete [] grad;
    } else if( x.optimize>4 )
    {
	// x.optimize number of bins
	double best_scale = scale, best_dist = dist;

	for(int scales=0; scales<x.optimize; scales++)
	{
	    // try new initial theta
	    scale = ((double)(scales)*(1.0/((double)x.optimize)));
	    // to try lower scales than initial scale, modulo 1
	    scale -= (double)floor(scale);

	    calculate_dist(x);

	    if( dist < best_dist )
	    {
		best_dist = dist;
		best_scale = scale;
	    }
	}

	scale = best_scale;
	dist = best_dist;
    }

    // check patch generated in c code is the same
    /*mexPrintf("C code\n");
      double * spatch;
      spatch = new double[x.patch_w * x.patch_w * 3];
      get_patch(x,x.S,3,spatch);
      for( int k=0; k<3; k++ )
      {
      for( int r=-x.patch_offset; r<=x.patch_offset; r++ )
      {
      for (int c=-x.patch_offset; c<=x.patch_offset; c++ )
      {
      mexPrintf("%f ",spatch[index3(r+x.patch_offset,c+x.patch_offset,k,x.patch_w,x.patch_w)]);
      }
      mexPrintf("\n");
      }
      mexPrintf("\n");
      }*/

}

void patch_sc_db::random(images const & x, double const alpha) {
    // random

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_scale = scale, best_dist = dist;

    // random search
    double search_radius = (double)std::max(x.TC,std::max(x.TR,std::max(x.SC,x.SR)));
    double search_radius_scale = 1.0;
    while( search_radius > 1.0 )
    {
	sr += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));
	sc += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));

	// if this points to a valid patch
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		else
		{
		    //random_parameters(x);
		    
		    // GENERALIZED PATCH MATCH
		    scale += search_radius_scale*((2.0*((double)rand()/(double)RAND_MAX))-1.0);
		    scale -= (double)floor(scale);
		}

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    scale = best_scale;
		    dist = best_dist;
		} else {
		    // if have found a better match, update best
		    best_sr = sr;
		    best_sc = sc;
		    best_s = s;
		    best_scale = scale;
		    best_dist = dist;
		}	
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		scale = best_scale;
		dist = best_dist;
	    }						
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    scale = best_scale;
	    dist = best_dist;
	}

	search_radius *= alpha;
	search_radius_scale *= (1.5*alpha);
    }
}

void patch_sc_db::consider(patch_sc_db const & p, images const & x) {
   // consider parameters based on parameters at another location

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_scale = scale, best_dist = dist;

    int p_location[3];
    double p_parameters[patch_sc_db::num_parameters+1];
    p.get_location(p_location);
    p.get_parameters(p_parameters);

    if( p_location != NULL && p_parameters != NULL )
    {
	// p_location and p_parameters assigned

	scale = p_parameters[3];
	sr = (int)(tr - p_location[0]) + (int)p_parameters[0];
	sc = (int)(tc - p_location[1]) + (int)p_parameters[1];

	// check transformed patch match from S is valid
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		
		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    scale = best_scale;
		    dist = best_dist;
		}
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		scale = best_scale;
		dist = best_dist;
	    }
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    scale = best_scale;
	    dist = best_dist;
	}
    } //else do nothing
}
