
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <math.h>

#include "mex.h"

#define PI 3.14159265

/**************************************************************************
 * patch_sim
 **************************************************************************/

void patch_sim_db::get_patch(images const & x, double * Simagedb, double * patch) const 
{
    // patch should be allocated length x.pN * 3

    // apply rotation, scale and offset to patch indices

    // theta is in range [0,8]
    double ct = cos((theta / 4.0) * PI);
    double st = sin((theta / 4.0) * PI);

    double * ptr_Patchind = x.patchind;
    double * ptr_rMat = x.rMat;
    double * ptr_cMat = x.cMat;

    double r, c;
    for( int j=0; j<x.pN; j++ )
    {
	r = *ptr_Patchind - (double)x.patchdbS_offset;
	c = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	*ptr_rMat = (1.0+scale)*(ct*r + st*c) + (double)x.patchdbS_offset + 1.0;
	*ptr_cMat = (1.0+scale)*(ct*c - st*r) + (double)x.patchdbS_offset + 1.0;

	ptr_rMat++;
	ptr_cMat++;
    }

    vgg_interp2_c_db(patch,&(Simagedb[index2(0,s,x.patchdbS_w*x.patchdbS_w*3)]),x.patchdbS_w,x.pN,x.cMat,x.rMat);

    //mexPrintf("(theta %f)",theta);

    return;
}

void patch_sim_db::assign_parameters(double * parameters) 
{
    // parameters should be allocated length patch_sim::num_parameters
    sr = (int)parameters[0];
    sc = (int)parameters[1];
    s = (int)parameters[2];
    theta = parameters[3];
    scale = parameters[4];
}

void patch_sim_db::get_parameters(double * parameters) const 
{
    // parameters should be allocated length patch_sim::num_parameters
    parameters[0] = (double)sr;
    parameters[1] = (double)sc;
    parameters[2] = (double)s;
    parameters[3] = theta;
    parameters[4] = scale;
}

void patch_sim_db::random_parameters(images const & x) 
{
    if( x.optimize>4 )
    {
	// random discrete theta
	theta = ((double)(rand() % x.optimize)*(8.0/((double)x.optimize)));
	// random discrete scale
	// 0<scale<=1.0
	scale = ((double)(rand() % x.optimize))*(1.0/((double)x.optimize));
    } else {
	// random continous 
	// theta is in range [0,8]
	// 0<scale<=1.0
	theta = 8.0*((double)rand()/(double)RAND_MAX);
	scale = ((double)rand()+1.0)/((double)RAND_MAX+1.0);
    }
}

void patch_sim_db::gradient_lm_rot(images const & x, double * err, double * grad) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    // err and grad should be allocated to length 3*pN

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    // determine rotation gradient matrix
    double ct = cos((theta / 4.0) * PI);
    double st = sin((theta / 4.0) * PI);

    double ind_x, ind_y, ind_x_grad, ind_y_grad;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	ind_x_grad = -st*ind_x - ct*ind_y;
	ind_y_grad = ct*ind_x - st*ind_y;
	
	for( int k=0; k<3; k++ )
	{
	    grad[index2(k,i,3)] = (1.0-*ptr_Tmask)*((*ptr_patchSgradx*ind_x_grad + *ptr_patchSgrady*ind_y_grad));
	    err[index2(k,i,3)] =  (1.0-*ptr_Tmask)*(*ptr_Tpatch - *ptr_Spatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	ptr_Tmask++;
    }
}

void patch_sim_db::gradient_lm_sc(images const & x, double * err, double * grad) const
{
    // THIS SHOULD ONLY BE CALLED FOR VALID SOURCE PATCHES AFTER CALLING CALCULATE_DIST WITH THE SAME PARAMETERS
    // (THIS SETS X.PATCHS CORRECTLY)
    // err and grad should be allocated to length 3*pN

    // get interpolated patch of Sgrad_x and Sgrad_y
    get_patch(x, x.Sgrad_xdb, x.patchSgradx);
    get_patch(x, x.Sgrad_ydb, x.patchSgrady);

    double ind_x, ind_y;
    
    // determine gradient
    double * ptr_Patchind = x.patchind;
    double * ptr_Tmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    double * ptr_Spatch = x.patchS;
    double * ptr_Tpatch = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_patchSgradx = x.patchSgradx;
    double * ptr_patchSgrady = x.patchSgrady;
    for(int i=0; i<x.pN; i++)
    {
	ind_y = *ptr_Patchind - (double)x.patchdbS_offset;
	ind_x = *(++ptr_Patchind) - (double)x.patchdbS_offset;
	ptr_Patchind++;

	for( int k=0; k<3; k++ )
	{
	    grad[index2(k,i,3)] = (1.0-*ptr_Tmask)*((*ptr_patchSgradx*ind_x + *ptr_patchSgrady*ind_y));
	    err[index2(k,i,3)] =  (1.0-*ptr_Tmask)*(*ptr_Tpatch - *ptr_Spatch);
	    ptr_Spatch++;
	    ptr_Tpatch++;
	    ptr_patchSgradx++;
	    ptr_patchSgrady++;
	}

	ptr_Tmask++;
    }
}

void patch_sim_db::optimize_parameters(images const & x) {
//     optimize theta and scale for current (sr,sc)

    if( x.optimize==4 )
    {
	// Levenburg-Marquadt

	// parameters
	int const no_theta = 6;
	int const no_scale = 6;
	double const eps_1 = 1e-5;
	double const eps_2 = 1e-5;
	double const eps_3 = 1e-5;
	double const tau = 1e-2;
	int const kmax = 100;

	// for storing the best
	double theta_best = theta;
	double scale_best = scale;
	double dist_best = DISTMAX;

	// for storing bounds
	double theta_max = 8.0;
	double theta_min = 0.0;

	double * err = new double[3*x.pN];
	double * grad_rot = new double[3*x.pN];
	double * grad_sc = new double[3*x.pN];

	double const initial_theta = theta;
	double const initial_scale = scale;
	double diveight = 0.0;
	for(int scales=0; scales<no_scale; scales++)
	{
	    for(int thetas=0; thetas<no_theta; thetas++)
	    {
		// try new initial theta
		theta = initial_theta + ((double)(thetas)*(8.0/((double)no_theta)));
		// having found theta, set up the bounds
		theta_max = theta + 4.0/(double)no_theta;
		theta_min = theta - 4.0/(double)no_theta;

		// try new initial scale
		scale = initial_scale + ((double)(scales)*(1.0/((double)no_scale)));
		// to try lower scales than initial scale, modulo 1
		scale -= (double)floor(scale);
	    
		double nu = 2.0;

		calculate_dist(x);
		gradient_lm_rot(x,err,grad_rot);
		gradient_lm_rot(x,err,grad_sc);

		double A_rot_rot = 0.0;
		double A_rot_sc = 0.0;
		double A_sc_sc = 0.0;
		double g_rot = 0.0;
		double g_sc = 0.0;
		double * ptr_err = err;
		double * ptr_grad_rot = grad_rot;
		double * ptr_grad_sc = grad_sc;
		for( int i=0; i<x.pN; i++ )
		{
		    for( int j=0; j<3; j++ )
		    {
			g_rot += (*ptr_grad_rot)*(*ptr_err);
			g_sc += (*ptr_grad_sc)*(*ptr_err);
			A_rot_rot += (*ptr_grad_rot)*(*ptr_grad_rot);
			A_rot_sc += (*ptr_grad_rot)*(*ptr_grad_sc);
			A_sc_sc += (*ptr_grad_sc)*(*ptr_grad_sc);
		    
			ptr_err++;
			ptr_grad_rot++;
			ptr_grad_sc++;
		    }
		}

		int stop = 0;
	
		if( std::max(fabs(g_rot),fabs(g_sc)) < eps_1 )
		    stop = 1;

		double mu = tau * std::max(A_rot_rot,A_sc_sc);

		int k = 0;
		while( stop==0 && (k<kmax) )
		{
		    k++;

		    double rho = 0.0;
		    while( stop==0 && rho<=0.0 )
		    {
			double delta_theta = 0.0;
			double delta_scale = 0.0;
			
			delta_scale = g_sc*(A_rot_rot+mu - A_rot_sc) / ((A_rot_rot+mu)*(A_sc_sc+mu) - A_rot_sc*A_rot_sc);
			delta_theta = (g_rot - A_rot_sc*delta_scale) / (A_rot_rot+mu);

			if( sqrt(delta_theta*delta_theta + delta_scale*delta_scale) <= eps_2 )
			{
			    stop = 1;
			} else {
			    double theta_old = theta;
			    double scale_old = scale;
			    theta += delta_theta;
			    //thresholds to bounds
			    if( (theta>=theta_max) || (theta<=theta_min) )
			    {
				theta = std::max(std::min(theta,theta_max),theta_min);
			    }
			    scale += delta_scale;
			    //thresholds to bounds
			    if( (scale>=0.0) || (scale<=1.0) )
			    {
				scale = std::max(std::min(scale,1.0),0.0);
			    }
			    // CAN WE STOP HERE? ON WHAT CONDITIONS?
			    
			    double dist_old = dist; 
			    calculate_dist(x);
			
			    rho = ((dist_old - dist)/(delta_theta*(mu*delta_theta + g_rot) + delta_scale*(mu*delta_scale + g_sc) ));
			    if( rho>0 )
			    {
				gradient_lm_rot(x,err,grad_rot);
				gradient_lm_sc(x,err,grad_sc);

				A_rot_rot = 0.0;
				A_rot_sc = 0.0;
				A_sc_sc = 0.0;
				g_rot = 0.0;
				g_sc = 0.0;
				ptr_err = err;
				ptr_grad_rot = grad_rot;
				ptr_grad_sc = grad_sc;
				for( int i=0; i<x.pN; i++ )
				{
				    for( int j=0; j<3; j++ )
				    {
					g_rot += (*ptr_grad_rot)*(*ptr_err);
					g_sc += (*ptr_grad_sc)*(*ptr_err);
					A_rot_rot += (*ptr_grad_rot)*(*ptr_grad_rot);
					A_rot_sc += (*ptr_grad_rot)*(*ptr_grad_sc);
					A_sc_sc += (*ptr_grad_sc)*(*ptr_grad_sc);
		    
					ptr_err++;
					ptr_grad_rot++;
					ptr_grad_sc++;
				    }
				}

				if( (std::max(fabs(g_rot),fabs(g_sc)) < eps_1) || (dist <= eps_3) )
				    stop = 1;

				mu *= std::max(1.0/3.0,1-(2.0*rho-1.0)*(2.0*rho-1.0)*(2.0*rho-1.0));
				nu = 2.0;
			
			    } else {
				theta = theta_old;
				scale = scale_old;
				dist = dist_old;

				mu *= nu;
				nu *= 2.0;
			    }
			}
		    }
		}

		if( dist < dist_best )
		{
		    dist_best = dist;
		    theta_best = theta;
		    scale_best = scale;
		}
	    }
	}

	dist = dist_best;
	diveight = theta_best/8.0;
	theta_best -= 8.0*(double)floor(diveight);
	theta = theta_best;
	scale = scale_best;

	delete [] err;
	delete [] grad_rot;
	delete [] grad_sc;
    } else if( x.optimize>4 )
    {
	// x.optimize number of bins
	double diveight = 0.0;
	double best_theta = theta, best_scale = scale, best_dist = dist;

	for(int thetas=0; thetas<x.optimize; thetas++)
	{
	    // try new initial theta
	    theta = ((double)(thetas)*(8.0/((double)x.optimize)));
	    for(int scales=0; scales<x.optimize; scales++)
	    {
		// try new initial theta
		scale = ((double)(scales)*(1.0/((double)x.optimize)));
		// to try lower scales than initial scale, modulo 1
		scale -= (double)floor(scale);

		calculate_dist(x);
		
		if( dist < best_dist )
		{
		    best_dist = dist;
		    best_theta = theta;
		    best_scale = scale;
		}
	    }
	}

	diveight = best_theta/8.0;
	best_theta -= 8.0*(double)floor(diveight);
	theta = best_theta;
	scale = best_scale;
	dist = best_dist;
    }


// check patch generated in c code is the same
/*mexPrintf("C code\n");
  double * spatch;
  spatch = new double[x.patch_w * x.patch_w * 3];
  get_patch(x,x.S,3,spatch);
  for( int k=0; k<3; k++ )
  {
  for( int r=-x.patch_offset; r<=x.patch_offset; r++ )
  {
  for (int c=-x.patch_offset; c<=x.patch_offset; c++ )
  {
  mexPrintf("%f ",spatch[index3(r+x.patch_offset,c+x.patch_offset,k,x.patch_w,x.patch_w)]);
  }
  mexPrintf("\n");
  }
  mexPrintf("\n");
  }*/

}

void patch_sim_db::random(images const & x, double const alpha) {
    // random

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_theta = theta, best_scale = scale, best_dist = dist;

    // random search
    double search_radius = (double)std::max(x.TC,std::max(x.TR,std::max(x.SC,x.SR)));
    double search_radius_scale = 1.0;
    double search_radius_theta = 8.0;
    while( search_radius > 1.0 )
    {
	sr += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));
	sc += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));

	// if this points to a valid patch
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		else
		{		
		    //random_parameters(x);
		    
		    // GENERALIZED PATCH MATCH
		    theta += search_radius_theta*((2.0*((double)rand()/(double)RAND_MAX))-1.0);
		    double diveight = theta/8.0;
		    theta -= 8.0*(double)floor(diveight);

		    scale += search_radius_scale*((2.0*((double)rand()/(double)RAND_MAX))-1.0);
		    scale -= (double)floor(scale);
		}

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    theta = best_theta;
		    scale = best_scale;
		    dist = best_dist;
		} else {
		    // if have found a better match, update best
		    best_sr = sr;
		    best_sc = sc;
		    best_s = s;
		    best_theta = theta;
		    best_scale = scale;
		    best_dist = dist;
		}	
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		theta = best_theta;
		scale = best_scale;
		dist = best_dist;
	    }						
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    theta = best_theta;
	    scale = best_scale;
	    dist = best_dist;
	}

	search_radius *= alpha;
	search_radius_theta *= (1.5*alpha);
	search_radius_scale *= (1.5*alpha);
    }
}

void patch_sim_db::consider(patch_sim_db const & p, images const & x) {
   // consider parameters based on parameters at another location

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_theta = theta, best_scale = scale, best_dist = dist;

    int p_location[3];
    double p_parameters[patch_sim_db::num_parameters+1];
    p.get_location(p_location);
    p.get_parameters(p_parameters);

    if( p_location != NULL && p_parameters != NULL )
    {
	// p_location and p_parameters assigned

	theta = p_parameters[3];
	scale = p_parameters[4];
	double ct = cos(theta / 4.0 * PI);
	double st = sin(theta / 4.0 * PI);
	sr = (int)(ct*(tr - p_location[0]) + st*(tc - p_location[1])) + (int)p_parameters[0];
	sc = (int)(ct*(tc - p_location[1]) - st*(tr - p_location[0])) + (int)p_parameters[1];

	// check transformed patch match from S is valid
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    theta = best_theta;
		    scale = best_scale;
		    dist = best_dist;
		}
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		theta = best_theta;
		scale = best_scale;
		dist = best_dist;	
	    }
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    theta = best_theta;
	    scale = best_scale;
	    dist = best_dist;
	}
    } //else do nothing
}
