
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

/*  transformic_pmnn.cpp
 *
 *  MEX function for Patch Match NN patch matching
 *  
 *  Based on description in the paper:
 *  'PatchMatch: A Randomized Correspondence Algorithm for Structural Image Editing'
 *  Barnes, Shechtman, Finkelstein and Goldman - SIGGRAPH 2009
 */ 

#include "transformic_pmnn.h"
#include "transformic_patchmatch_oo_db.h"
#include "transformic_exactmatch_oo_db.h"
#include "mex.h"

void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[]) 
{
    char strUsage[] = "Usage: \n\t nnfield = transformic_pmnn_oo(S,Sgrad_x,Sgrad_y,T,patch_w,Tmask,Sweight,Tweight,Tdomain,transformation,exact,optimize,use_nnfield0,nnfield0,use_patchdb,patchind,Sind,Tind,Sdb,Sindimage,Tindimage)\n\
         For each patch in T, finds NN patches in S using the Patch Match algorithm.\n\
	\t\t 0  S:  source image\n\
        \t\t 1  Sgrad_x: gradient of source image in x direction\n\
        \t\t 2  Sgrad_y gradient of source image in y direction\n\
    	\t\t 3  T: target image\n\
	\t\t 4  patch_w:  patch size\n\
    	\t\t 5  Tmask:  0, or 1 for pixels in T not to use in the matching score\n \
    	\t\t 6  Sweight:  weighting for patches in S\n\
    	\t\t 7  Tweight:  weighting for patches in T\n\
    	\t\t 8  Tdomain:  0, or 1 for patches in T for which to calculate the nnfield (returns default nnfield of direct copy)\n\
        \t\t 9  transformation: 0 for translation, 1 rotation, 2 scaling, 3 rotation+scaling, 4 vertical flipping, 5 illumination, 6 all\n\
        \t\t 10 exact: 0 for patchmatch algorithm, 1 to try all positions\n\
        \t\t 11 optimize: 0 for no, 1 for lsqnonlin, 2 with multiple initializations, 3 gradient descent, 4 our LM, >4 specifies number of uniformly spaced bins\n\
        \t\t 12 use_nnfield0: 0 for no, 1 for yes\n\
    	\t\t 13 nnfield0:  previous nnfield to consider (tested for closeness after a few random iterations)\n\
        \t\t 14 use_patchdb: 0 for no, 1 for yes \n\
        \t\t 15 patchind: patch indices 2 x pN\n\
        \t\t 16 Sind: source patch indices 3 x SN\n\
        \t\t 17 Tind: target patch indices 3 x TN\n\
        \t\t 18 Sdb: source patch database 3*(patchdbS_w+2)^2 x SN\n\
        \t\t 19 Sindimage: source patch indices over the image SR x SC\n\
        \t\t 20 Tindimage: target patch indices over the image TR x TC\n\
        \t\t 21 Sgrad_xdb: gradient of source image in x direction patch database\n\
        \t\t 22 Sgrad_ydb: gradient of source image in y direction patch database\n\
        \t\t 23 no_iterations: number of iterations of patch match\n";

    try
    {
	// check number of input/output arguments
	if(nrhs < 21)
	    throw("Needs 21 input arguments S, Sgrad_x, Sgrad_y, T, patch_w, Tmask, Sweight, Tweight, Tdomain, transformation, exact, optimize, use_nnfield0, nnfield0, use_patchdb, patchind, Sind, Tind, Sdb, Sindimage and Tindimage\n");
	else if(nrhs > 24)
	    throw("Needs no more than 23 input arguments S, Sgrad_x, Sgrad_y, T, patch_w, Tmask, Sweight, Tweight, Tdomain, transformation, exact, optimize, use_nnfield0, nnfield0, use_patchdb, patchind, Sind, Tind, Sdb, Sindimage, Tindimage, Sgrad_xdb, Sgrad_ydb and no_iterations\n");
	if(nlhs != 1)
	    throw("Needs 1 output argument nnfield\n");

	// check number of dimensions
	int const dimsS = mxGetNumberOfDimensions(prhs[0]);
	if(dimsS != 3)
	    throw("S should have 3 dimensions\n");
	int const dimsSgrad_x = mxGetNumberOfDimensions(prhs[1]);
	if(dimsSgrad_x != 3)
	    throw("Sgrad_x should have 3 dimensions\n");
	int const dimsSgrad_y = mxGetNumberOfDimensions(prhs[2]);
	if(dimsSgrad_y != 3)
	    throw("Sgrad_y should have 3 dimensions\n");
	int const dimsT = mxGetNumberOfDimensions(prhs[3]);
	if(dimsT != 3)
	    throw("T should have 3 dimensions\n");
	int const dimsPatch = mxGetNumberOfDimensions(prhs[4]);
	if(dimsPatch != 2)
	    throw("patch_w should have no more than 2 dimensions\n");
	int const dimsTmask = mxGetNumberOfDimensions(prhs[5]);
	if(dimsTmask != 2)
	    throw("Tmask should have 2 dimensions\n");
	int const dimsSweight = mxGetNumberOfDimensions(prhs[6]);
	if(dimsSweight != 2)
	    throw("Sweight should have 2 dimensions\n");
	int const dimsTweight = mxGetNumberOfDimensions(prhs[7]);
	if(dimsTweight != 2)
	    throw("Tweight should have 2 dimensions\n");
	int const dimsTdomain = mxGetNumberOfDimensions(prhs[8]);
	if(dimsTmask != 2)
	    throw("Tdomain should have 2 dimensions\n");
	int const dimsTrans = mxGetNumberOfDimensions(prhs[9]);
	if(dimsTrans > 2)
		throw("transformation should have no more than 2 dimensions\n");
	int const dimsExact = mxGetNumberOfDimensions(prhs[10]);
	if(dimsExact > 2)
	    throw("exact should have no more than 2 dimensions\n");
	int const dimsOptimize = mxGetNumberOfDimensions(prhs[11]);
	if(dimsOptimize > 2)
	    throw("optimize should have no more than 2 dimensions\n");
	int const dimsUse_NN0 = mxGetNumberOfDimensions(prhs[12]);
	if(dimsUse_NN0 > 2)
	    throw("use_nnfield0 should have no more than 2 dimensions\n");
	int const dimsNN0 = mxGetNumberOfDimensions(prhs[13]);
	//if(dimsNN0 != 3)
	//    throw("nnfield0 should have 3 dimensions\n");
	int const dimsUse_patchdb = mxGetNumberOfDimensions(prhs[14]);
	if(dimsUse_patchdb > 2)
	    throw("use_patchdb should have no more than 2 dimensions\n");
	int const dimsPatchind = mxGetNumberOfDimensions(prhs[15]);
	//if(dimsPatchind != 2)
	//    throw("patchind should have 2 dimensions\n");
	int const dimsSind = mxGetNumberOfDimensions(prhs[16]);
	//if(dimsSind != 2)
	//    throw("Sind should have 2 dimensions\n");
	int const dimsTind = mxGetNumberOfDimensions(prhs[17]);
	//if(dimsTind != 2)
	//    throw("Tind should have 2 dimensions\n");
	int const dimsSdb = mxGetNumberOfDimensions(prhs[18]);
	//if(dimsSdb != 2)
	//    throw("Sdb should have 2 dimensions\n");
	int const dimsSindimage = mxGetNumberOfDimensions(prhs[19]);
	int const dimsTindimage = mxGetNumberOfDimensions(prhs[20]);
	int dimsSgrad_xdb = 0;
	if( nrhs>=25 )
	    dimsSgrad_xdb = mxGetNumberOfDimensions(prhs[21]);
	int dimsSgrad_ydb = 0;
	if( nrhs>=26 )
	    dimsSgrad_ydb = mxGetNumberOfDimensions(prhs[22]);
	int dimsNo_iterations = 0;
	if( nrhs>=27 )
	    dimsNo_iterations = mxGetNumberOfDimensions(prhs[23]);

	// check dimensions
	mwSize const * const sizeS = mxGetDimensions(prhs[0]);
	mwSize const * const sizeSgrad_x = mxGetDimensions(prhs[1]);
	if((sizeSgrad_x[0]!=sizeS[0])||(sizeSgrad_x[1]!=sizeS[1]))
	    throw("S and Sgrad_x should be the same size\n");
	mwSize const * const sizeSgrad_y = mxGetDimensions(prhs[2]);
	if((sizeSgrad_y[0]!=sizeS[0])||(sizeSgrad_y[1]!=sizeS[1]))
	    throw("S and Sgrad_y should be the same size\n");
	mwSize const * const sizeT = mxGetDimensions(prhs[3]);
	mwSize const * const sizePatch = mxGetDimensions(prhs[4]);
	if((sizePatch[0]!=1)||(sizePatch[1]!=1))
	    throw("patch_w should be scalar\n");
	mwSize const * const sizeTmask = mxGetDimensions(prhs[5]);
	if((sizeTmask[0]!=sizeT[0])||(sizeTmask[1]!=sizeT[1]))
	    throw("T and Tmask should be the same size\n");
	mwSize const * const sizeSweight = mxGetDimensions(prhs[6]);
	if((sizeSweight[0]!=sizeS[0])||(sizeSweight[1]!=sizeS[1]))
	    throw("S and Sweight should be the same size\n");
	mwSize const * const sizeTweight = mxGetDimensions(prhs[7]);
	if((sizeTweight[0]!=sizeT[0])||(sizeTweight[1]!=sizeT[1]))
	    throw("T and Tweight should be the same size\n");
	mwSize const * const sizeTdomain = mxGetDimensions(prhs[8]);
	if((sizeTdomain[0]!=sizeT[0])||(sizeTdomain[1]!=sizeT[1]))
	    throw("T and Tdomain should be the same size\n");
	mwSize const * const sizeTrans = mxGetDimensions(prhs[9]);
	if((sizeTrans[0]!=1)||(sizeTrans[1]!=1))
	    throw("transformation should be scalar\n");
	mwSize const * const sizeExact = mxGetDimensions(prhs[10]);
	if((sizeExact[0]!=1)||(sizeExact[1]!=1))
	    throw("exact should be scalar\n");
	mwSize const * const sizeOptimize = mxGetDimensions(prhs[11]);
	if((sizeOptimize[0]!=1)||(sizeOptimize[1]!=1))
	    throw("optimize should be scalar\n");
	mwSize const * const sizeUse_NN0 = mxGetDimensions(prhs[12]);
	if((sizeUse_NN0[0]!=1)||(sizeUse_NN0[1]!=1))
	    throw("use_nnfield0 should be scalar\n");
	mwSize const * const sizeNN0 = mxGetDimensions(prhs[13]);
	//if((sizeNN0[0]!=sizeT[0])||(sizeNN0[1]!=sizeT[1]))
	//    throw("T and nnfield0 should be the same size\n");
	mwSize const * const sizeUse_patchdb = mxGetDimensions(prhs[14]);
	if((sizeUse_patchdb[0]!=1)||(sizeUse_patchdb[1]!=1))
	    throw("patchdb should be scalar\n");
	mwSize const * const sizePatchind = mxGetDimensions(prhs[15]);
	//if(sizePatchind[0]!=2)
	//    throw("patchind should be 2 x pN");
	mwSize const * const sizeSind = mxGetDimensions(prhs[16]);
	//if(sizeSind[0]!=3)
	//    throw("Sind should be 3 x SN");
	mwSize const * const sizeTind = mxGetDimensions(prhs[17]);
	//if(sizeTind[0]!=3)
	//    throw("Tind should be 3 x TN");
	mwSize const * const sizeSdb = mxGetDimensions(prhs[18]);
	//if(sizeSdb[1]!=sizeSind[1])
	//    throw("Sdb should be 3*patchdbS_w^2 x SN");
	mwSize const * const sizeSindimage = mxGetDimensions(prhs[19]);
	mwSize const * const sizeTindimage = mxGetDimensions(prhs[20]);
	mwSize const * sizeSgrad_xdb = NULL;
	if( nrhs>=25 )
	    sizeSgrad_xdb = mxGetDimensions(prhs[21]);
	mwSize const * sizeSgrad_ydb = NULL;
	if( nrhs>=26 )
	    sizeSgrad_ydb = mxGetDimensions(prhs[22]);
	mwSize const * sizeNo_iterations = NULL;
	if( nrhs>=27 )
	    sizeNo_iterations = mxGetDimensions(prhs[23]);

	images x;

	// get values
	x.SR = (int)sizeS[0];
	x.SC = (int)sizeS[1];
	x.TR = (int)sizeT[0];
	x.TC = (int)sizeT[1];
	x.patch_w = (int)(*mxGetPr(prhs[4]));
	x.patch_offset = (int)(((float)(x.patch_w-1)) / 2.0);
	x.patchdbT_w = x.patch_w;
	x.patchdbT_offset = x.patch_offset;
	x.patchdbS_w = (int)sqrt((double)sizeSdb[0]/3.0);
	x.patchdbS_offset = (int)(((float)(x.patchdbS_w-1)) / 2.0);
	x.SN = (int)sizeSind[1];
	x.TN = (int)sizeTind[1];
	x.pN = (int)sizePatchind[1];

	int patch_pixels = x.patch_w*x.patch_w;
	
	// parameters
	int const iterations = (int)(*mxGetPr(prhs[23])); // number of iterations of propagation/random search                    
	int const iterations_before_previous = 2;	  // number of iterations before use nnfield0
	double const alpha = 0.5;
	srand(0);

	// are the source and target equal size?
	int equal_size = 0;
	if( (x.SR==x.TR) && (x.SC==x.TC) )
	    equal_size = 1;

	// get input pointers
	x.S = mxGetPr(prhs[0]);
	x.Sgrad_x = mxGetPr(prhs[1]);
	x.Sgrad_y = mxGetPr(prhs[2]);
	x.T = mxGetPr(prhs[3]);
	x.Tmask = mxGetPr(prhs[5]);
	x.Sweight = mxGetPr(prhs[6]);
	x.Tweight = mxGetPr(prhs[7]);
	x.Tdomain = mxGetPr(prhs[8]);
	int const transformation = (int)(*mxGetPr(prhs[9]));
	int const exact = (int)(*mxGetPr(prhs[10]));
	x.optimize = (int)(*mxGetPr(prhs[11]));
	int const use_nnfield0 = (int)(*mxGetPr(prhs[12]));
	x.nnfield0 = NULL;
	if( use_nnfield0>0 )
	    x.nnfield0 = mxGetPr(prhs[13]);
	int const use_patchdb = (int)(*mxGetPr(prhs[14]));

	if( (transformation>0)&&(use_patchdb>0)&&(nrhs<23) )
	    throw("need Sgrad_xdb and Sgrad_ydb when having translation above rotation and using patch database");

	x.patchind = NULL;
	x.Sind = NULL;
	x.Tind = NULL;
	x.Sdb = NULL;
	x.Sgrad_xdb = NULL;
	x.Sgrad_ydb = NULL;
	x.Sindimage = NULL;
	x.Tindimage = NULL;
	if( use_patchdb>0 )
	{
	    x.patchind = mxGetPr(prhs[15]);
	    x.Sind = mxGetPr(prhs[16]);
	    x.Tind = mxGetPr(prhs[17]);
	    x.Sdb = mxGetPr(prhs[18]);
	    if( transformation>0 )
	    {
		x.Sgrad_xdb = mxGetPr(prhs[21]);
		x.Sgrad_ydb = mxGetPr(prhs[22]);
	    }
	    x.Sindimage = mxGetPr(prhs[19]);
	    x.Tindimage = mxGetPr(prhs[20]);
	    patch_pixels = x.pN;
	}

	// define working arrays
	x.rMat = new double[patch_pixels];
	x.cMat = new double[patch_pixels];
	//mwSize sizeWorking[3];
	//sizeWorking[0] = (mwSize)x.SR;
	//sizeWorking[1] = (mwSize)x.SC;
	//sizeWorking[2] = (mwSize)3;
	//x.mxS = mxCreateNumericArray((mwSize)3,sizeWorking,mxDOUBLE_CLASS,mxREAL);
	//x.mxSgrad_x = mxCreateNumericArray((mwSize)3,sizeWorking,mxDOUBLE_CLASS,mxREAL);
	//x.mxSgrad_y = mxCreateNumericArray((mwSize)3,sizeWorking,mxDOUBLE_CLASS,mxREAL);
	x.mxTpatch = mxCreateDoubleMatrix((mwSize)(3*patch_pixels),(mwSize)1,mxREAL);
	x.mxTmask = mxCreateDoubleMatrix((mwSize)(3*patch_pixels),(mwSize)1,mxREAL);
	x.mxIxpatch = mxCreateDoubleMatrix((mwSize)(patch_pixels),(mwSize)2,mxREAL);
	//x.ptr_S = mxGetPr(x.mxS);
	//x.ptr_Sgrad_x = mxGetPr(x.mxSgrad_x);
	//x.ptr_Sgrad_y = mxGetPr(x.mxSgrad_y);
	x.ptr_Tpatch = mxGetPr(x.mxTpatch);
	x.ptr_Tmask = mxGetPr(x.mxTmask);
	x.ptr_Ixpatch = mxGetPr(x.mxIxpatch);

	//memcpy(x.ptr_S,x.S,x.SR*x.SC*3*sizeof(double));
	//memcpy(x.ptr_Sgrad_x,x.Sgrad_x,x.SR*x.SC*3*sizeof(double));
	//memcpy(x.ptr_Sgrad_y,x.Sgrad_y,x.SR*x.SC*3*sizeof(double));

	if( !(use_patchdb>0) )
	{
	    // as x.patchind already contains what we need for Ixpatch
	    for( int c=-x.patch_offset; c<=x.patch_offset; c++ )
	    {
		for (int r=-x.patch_offset; r<=x.patch_offset; r++ )
		{
		    x.ptr_Ixpatch[index2((r+x.patch_offset)+(c+x.patch_offset)*x.patch_w,1,x.patch_w*x.patch_w)] = r;
		    x.ptr_Ixpatch[index2((r+x.patch_offset)+(c+x.patch_offset)*x.patch_w,0,x.patch_w*x.patch_w)] = c;
		}
	    }
	}

	x.patchS = new double[patch_pixels*3];
	if(x.patchS == NULL)
	    throw("Could not allocate x.patchS\n");
	x.patchSmask = new double[patch_pixels];
	if(x.patchSmask == NULL)
	    throw("Could not allocate x.patchSmask\n");
	x.patchSgradx = new double[patch_pixels*3];
	if(x.patchSgradx == NULL)
	    throw("Could not allocate x.patchSgradx\n");
	x.patchSgrady = new double[patch_pixels*3];
	if(x.patchSgrady == NULL)
	    throw("Could not allocate x.patchSgrady\n");

	// target patch database
	x.Tdb = NULL;
	x.Tmaskdb = NULL;
	if ( use_patchdb>0 )
	{
	    x.Tdb = new double[3*x.pN*x.TN];
	    x.Tmaskdb = new double[x.pN*x.TN];

	    double * ptr_Tind = x.Tind, * ptr_Tdb = x.Tdb, * ptr_Tmaskdb = x.Tmaskdb;
	    int r, c, tr, tc;
	    for( int i=0; i<x.TN; i++ )
	    {
		r = (int)*ptr_Tind;
		c = (int)*(++ptr_Tind);
		ptr_Tind++;
		ptr_Tind++;

// 		mexPrintf("t %i r %i c %i ",i,r,c);
		
		double * ptr_Patchind = x.patchind;
		for( int j=0; j<x.pN; j++ )
		{
		    tr = r+(int)*ptr_Patchind-x.patchdbS_offset;
		    tc = c+(int)*(++ptr_Patchind)-x.patchdbS_offset;
		    for( int k=0; k<3; k++ )
		    {
			*ptr_Tdb = x.T[index3(tr,tc,k,x.TR,x.TC)];
// 			mexPrintf("tr%i tc%i %f ",tr,tc,*ptr_Tdb);
			ptr_Tdb++;
		    }
		    *ptr_Tmaskdb = x.Tmask[index2(tr,tc,x.TR)];
		    ptr_Tmaskdb++;
		    
		    ptr_Patchind++;
		}
// 		mexPrintf(")\n");
	    }
	}
       
	// define output arrays
	mwSize sizeOut[3];
	sizeOut[0] = (mwSize)x.TR;
	sizeOut[1] = (mwSize)x.TC;
	
	if( transformation==1 )
	    sizeOut[2] = (mwSize)(1 + patch_rot_db::num_parameters);
	else if( transformation==2 )
	    sizeOut[2] = (mwSize)(1 + patch_sc_db::num_parameters);
	else if( transformation==3 )
	    sizeOut[2] = (mwSize)(1 + patch_sim_db::num_parameters);
	else if( transformation==4 )
	    sizeOut[2] = (mwSize)(1 + patch_fl_db::num_parameters);
	else if( transformation==5 )
	    sizeOut[2] = (mwSize)(1 + patch_il_db::num_parameters);
	else if( transformation==6 )
	    sizeOut[2] = (mwSize)(1 + patch_simb_db::num_parameters);
	else 
	    sizeOut[2] = (mwSize)(1 + patch_tr_db::num_parameters); // transformation 0

	plhs[0] = mxCreateNumericArray(3,sizeOut,mxDOUBLE_CLASS,mxREAL); //initializes to 0
	if(plhs[0] == NULL)
	    throw("Could not allocate output\n");
	double * nnfield = mxGetPr(plhs[0]);

	if ( use_patchdb>0 )
	{
	    if( exact==0 )
	    {
		if( transformation==1 )
		    transformic_patchMatch_db<patch_rot_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==2 )
		    transformic_patchMatch_db<patch_sc_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==3 )
		    transformic_patchMatch_db<patch_sim_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==4 )
		    transformic_patchMatch_db<patch_fl_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==5 )
		    transformic_patchMatch_db<patch_il_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==6 )
		    transformic_patchMatch_db<patch_simb_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else 
		    transformic_patchMatch_db<patch_tr_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0); // transformation 0
	    } else {
		if( transformation==1 )
		    transformic_exactMatch_db<patch_rot_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==2 )
		    transformic_exactMatch_db<patch_sc_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==3 )
		    transformic_exactMatch_db<patch_sim_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==4 )
		    transformic_exactMatch_db<patch_fl_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==5 )
		    transformic_exactMatch_db<patch_il_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else if( transformation==6 )
		    transformic_exactMatch_db<patch_simb_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0);
		else 
		    transformic_exactMatch_db<patch_tr_db>(nnfield,x,iterations,iterations_before_previous,alpha,use_nnfield0); // transformation 0
	    }
	}
	
	delete [] x.rMat;
	delete [] x.cMat;
	delete [] x.patchS;
	delete [] x.patchSmask;
	if ( use_patchdb>0 )
	{
	    delete [] x.Tdb;
	    delete [] x.Tmaskdb;
	}

	return;
    } 
    catch(std::bad_alloc&)
    {
	mexErrMsgTxt("Out of memory");
    }
    catch(const char* p)
    {
	std::string strOut(p);
	strOut += "\n";
	strOut += strUsage;
	mexErrMsgTxt(strOut.c_str());
    }
    catch(...)
    {
	mexErrMsgTxt("Unknown exception caught");
    }

}

