
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#ifndef TRANSFORMIC_PMNN_H
#define TRANSFORMIC_PMNN_H

/*
 To add new transformations:
 - Add to transformic_pmnn.h
 - Create transformic_patchmatch_patchXX.cpp
 - Create transformic_patchmatch_patchXX_db.cpp
 - Add these files to the Makefile
 - Add condition to:
   - transformic_pmnn.cpp (NN call and nnfield size)
   - transformic_vote_XXXX.cpp
   - transformic_get_patch.cpp
 - Set appropriate patch indices for database in transformic_inpaint.m
 - Set templates to compile in transformic_*match_oo*.h
 */

#ifdef _WIN32
// these lines added to prevent conflict in definitions of char16_t
#include <yvals.h> 
#if (_MSC_VER >= 1600)
#define __STDC_UTF_16__
#endif
#endif

#include <stdexcept>
#include <string>
#include <cmath>
#include <algorithm>
#include <float.h>

#include "mex.h"

#define DISTMAX 1000000000;

inline int index2(int r, int c, int R) {
    return c*R + r;
}

inline int index3(int r, int c, int t, int R, int C) {
    return (t*C + c)*R + r;
}

struct images {
    double * S;
    double * Sgrad_x;
    double * Sgrad_y;
    double * T;
    double * Smask;
    double * Tmask;
    double * SsegMask;
    double * TsegMask;
    double * Sweight;
    double * Tweight;
    double * Tdomain;
    double * nnfield0;
    int TR, TC, SR, SC, patch_w, patch_offset, optimize;

    double * rMat, * cMat;
    mxArray * mxS, * mxSgrad_x, * mxSgrad_y, * mxTpatch, * mxTmask, * mxIxpatch;
    //double * ptr_S, * ptr_Sgrad_x, * ptr_Sgrad_y;
    double * ptr_Tpatch, * ptr_Tmask, * ptr_Ixpatch;

    double * patchS, * patchSmask, * patchSsegMask, * patchSgradx, *patchSgrady;

    // patch database structures
    int SN, TN, pN, patchdbT_w, patchdbT_offset, patchdbS_w, patchdbS_offset;
    double * Sind, * Tind, * patchind, * Sindimage, * Tindimage, * Sdb, * Sgrad_xdb, * Sgrad_ydb;
    double * Tdb, * Tmaskdb;
};

class patch_tr_db {
public:
patch_tr_db() : calls(0), tr(0), tc(0), sr(0), sc(0), dist(DBL_MAX), t(0), s(0) { dist = DISTMAX; }
    virtual ~patch_tr_db() {}

    virtual void assign_location(int r, int c, int i) {t = i; tr = r; tc = c;}
    virtual void assign_parameters(double * parameters);

    virtual void get_location(int * location) const;
	double get_dist() const { return dist; };

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x) {};
    virtual void optimize_parameters(images const & x) {};
    
    virtual void initialize(images const & x);
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_tr_db const & p, images const & x);

    static int const num_parameters = 2;

	 // for debugging
    virtual void call_calculate_dist(images const & x) { calculate_dist(x); }
    int calls;
protected:
	// location
    int tr, tc;
    // parameters
    int sr, sc;
    // matching cost
    double dist;  
    // location index
    int t, s;

    virtual void calculate_dist(images const & x);
};

class patch_fl_db : public patch_tr_db {
public:
patch_fl_db() : orient(0) {}
    virtual ~patch_fl_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_fl_db const & p, images const & x);

    static int const num_parameters = 3;
protected:
    // parameters
    int orient;
};

class patch_il_db : public patch_tr_db {
public:
patch_il_db() : illum(0.0) {}
    virtual ~patch_il_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_il_db const & p, images const & x);

    static int const num_parameters = 3;
protected:
    // parameters
    double illum;
};

class patch_rot_db : public patch_tr_db {
public:
patch_rot_db() : theta(0.0) {}
    virtual ~patch_rot_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_rot_db const & p, images const & x);

    static int const num_parameters = 3;
protected:
    // parameters
    double theta;

    int gradient(images const & x) const;
    void gradient_lm(images const & x, double * err, double * grad) const;
};

class patch_sc_db : public patch_tr_db {
public:
patch_sc_db() : scale(0.0) {}
    virtual ~patch_sc_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_sc_db const & p, images const & x);

    static int const num_parameters = 3;
protected:
    // parameters
    double scale;

    int gradient(images const & x) const;
    void gradient_lm(images const & x, double * err, double * grad) const;
};

class patch_sim_db : public patch_tr_db {
public:
patch_sim_db() : theta(0.0), scale(0.0) {}
    virtual ~patch_sim_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_sim_db const & p, images const & x);

    static int const num_parameters = 4;
protected:
    // parameters
    double theta, scale;

    void gradient_lm_rot(images const & x, double * err, double * grad) const;
    void gradient_lm_sc(images const & x, double * err, double * grad) const;
};

class patch_simb_db : public patch_tr_db {
public:
patch_simb_db() : theta(0.0), scale(0.0), illum(0.0) {}
    virtual ~patch_simb_db() {}

    virtual void assign_parameters(double * parameters);

    virtual void get_patch(images const & x, double * Simagedb, double * patch) const;
    virtual void get_parameters(double * parameters) const;
    virtual void random_parameters(images const & x);
    virtual void optimize_parameters(images const & x);
    
    virtual void random(images const & x, double const alpha);
    virtual void consider(patch_simb_db const & p, images const & x);

    static int const num_parameters = 5;
protected:
    // parameters
    double theta, scale, illum;

    void gradient_lm_rot(images const & x, double * err, double * grad) const;
    void gradient_lm_sc(images const & x, double * err, double * grad) const;
    void optimize_illum(images const & x);
};

void vgg_interp2_c(int K, double *plhs, int const nrhs, const mxArray *prhs[], int SC, int SR, double * Simage, int patch_w, double * cMat, double * rMat);
void vgg_interp2_c_db(double * const B, double const * const A, int const patch_w, int const num_points, double const * const X, double const * const Y);

#endif
