
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

/*  transformic_vote_weightedmedian.cpp
 *
 *  MEX function for pixel estimation from nnfield
 *
 *  Alex Mansfield 2011, alex.mansfield@vision.ee.ethz.ch
 */ 

#ifdef _WIN32
// these lines added to prevent conflict in definitions of char16_t
#include <yvals.h> 
#if (_MSC_VER >= 1600)
#define __STDC_UTF_16__
#endif
#endif

#include "mex.h"
#include <stdexcept>
#include <string>
#include <cmath>
#include <list>

#include "transformic_pmnn.h"
#include "transformic_vote_weightedmedian_db.h"

bool compare_votes(std::pair<double,double> first, std::pair<double,double> second)
{
    bool retval = false;

    if( first.first < second.first )
	retval = true;

    return retval;
}

void mexFunction(int nlhs, mxArray * plhs[], int nrhs, const mxArray * prhs[]) 
{

    char strUsage[] = "Usage: \n\t [T Tweight] = transformic_vote_weightedmedian(S,nnfield,patch_w,Tweight,Tdomain,transformation)\n\
	For each patch in T, votes for colours of pixels in T using NN matching patches in S.\n\
	These pixel colour is chosen by the median of these votes.\n\
	NOTE: This requires a lot of memory!\n\
	\t\t S:  source image\n\
	\t\t nnfield:  nearest neighbour field\n\
	\t\t patch_w:  patch size\n\
	\t\t Tweight:  weight of votes based on confidence in patches in T used to create nnfield\n\
        \t\t Tdomain: 0, or 1 for patches in T for which to perform voting\n	\
        \t\t transformation: 0 for translation, 1 for rotation, 2 for similarity\n";

    try
    {
	// check number of input/output arguments
	if(nrhs != 10)
	    throw("Needs 10 input arguments S, nnfield, patch_w, Tweight, Tdomain, transformation, use_patchdb, patchind, Sdb and Sindimage\n");
	if(nlhs > 2)
	    throw("Takes at most 2 output arguments T and Tmask\n");

	// check number of dimensions
	const int dimsS = mxGetNumberOfDimensions(prhs[0]);
	if(dimsS != 3)
	    throw("S should have 3 dimensions\n");
	const int dimsNN = mxGetNumberOfDimensions(prhs[1]);
	if(dimsNN != 3)
	    throw("nnfield should have 3 dimensions\n");
	const int dimsPatch = mxGetNumberOfDimensions(prhs[2]);
	if(dimsPatch != 2)
	    throw("patch_w should have no more than 2 dimensions\n");
	const int dimsTweight = mxGetNumberOfDimensions(prhs[3]);
	if(dimsTweight != 2)
	    throw("Tweight should have 2 dimensions\n");
	const int dimsTdomain = mxGetNumberOfDimensions(prhs[4]);
	if(dimsTdomain != 2)
	    throw("Tdomain should have 2 dimensions\n");
	int dimsTrans = mxGetNumberOfDimensions(prhs[5]);
	//if(dimsTrans > 2)
	//    throw("transformation should have no more than 2 dimensions\n");
	int dimsUse_patchdb = mxGetNumberOfDimensions(prhs[6]);
	//if(dimsUse_patchdb > 2)
	//    throw("use_patchdb should have no more than 2 dimensions\n");
	int const dimsPatchind = mxGetNumberOfDimensions(prhs[7]);
	//if(dimsPatchind != 2)
	//    throw("patchind should have 2 dimensions\n");
	int const dimsSdb = mxGetNumberOfDimensions(prhs[8]);
	//if(dimsSdb != 2)
	//    throw("Sdb should have 2 dimensions\n");
	int const dimsSindimage = mxGetNumberOfDimensions(prhs[9]);

	// check dimensions
	const int * sizeS = mxGetDimensions(prhs[0]);
	const int * sizeNN = mxGetDimensions(prhs[1]);
	const int * sizePatch = mxGetDimensions(prhs[2]);
	if((sizePatch[0]!=1)||(sizePatch[1]!=1))
	    throw("patch_w should be scalar\n");
	const int * sizeTweight = mxGetDimensions(prhs[3]);
	if((sizeNN[0]!=sizeTweight[0])||(sizeNN[1]!=sizeTweight[1]))
	    throw("nnfield and Tweight should have the same number of rows and columns\n");
	const int * sizeTdomain = mxGetDimensions(prhs[4]);
	if((sizeNN[0]!=sizeTdomain[0])||(sizeNN[1]!=sizeTdomain[1]))
	    throw("nnfield and Tdomain should have the same number of rows and columns\n");
	mwSize const * const sizeTrans = mxGetDimensions(prhs[5]);
	//if((sizeTrans[0]!=1)||(sizeTrans[1]!=1))
	//    throw("transformation should be scalar\n");
	mwSize const * const sizeUse_patchdb = mxGetDimensions(prhs[6]);
	//if((sizeUse_patchdb[0]!=1)||(sizeUse_patchdb[1]!=1))
	//    throw("patchdb should be scalar\n");
	mwSize const * const sizePatchind = mxGetDimensions(prhs[7]);
	//if(sizePatchind[0]!=2)
	//    throw("patchind should be 2 x pN");
	mwSize const * const sizeSdb = mxGetDimensions(prhs[8]);
	//if(sizeSdb[1]!=sizeSind[1])
	//    throw("Sdb should be 3*patchdbS_w^2 x SN");
	mwSize const * const sizeSindimage = mxGetDimensions(prhs[9]);

	images x;

	// get values
	x.SR = sizeS[0];
	x.SC = sizeS[1];
	x.TR = sizeNN[0];
	x.TC = sizeNN[1];
	x.patch_w = (int)(*mxGetPr(prhs[2]));
	x.patch_offset = (int)(((float)(x.patch_w-1)) / 2.0);
	x.patchdbT_w = x.patch_w;
	x.patchdbT_offset = x.patch_offset;
	x.patchdbS_w = (int)sqrt((double)sizeSdb[0]/3.0);
	x.patchdbS_offset = (int)(((float)(x.patchdbS_w-1)) / 2.0);
	x.pN = (int)sizePatchind[1];

	int patch_pixels = x.patch_w*x.patch_w;

	// get input pointers
	x.S = mxGetPr(prhs[0]);
	double * nnfield = mxGetPr(prhs[1]);
	x.Tweight = mxGetPr(prhs[3]);
	x.Tdomain = mxGetPr(prhs[4]);
	int transformation = (int)(*mxGetPr(prhs[5]));
	int const use_patchdb = (int)(*mxGetPr(prhs[6]));

	x.patchind = NULL;
	x.Sdb = NULL;
	x.Sindimage = NULL;
	if( use_patchdb>0 )
	{
	    x.patchind = mxGetPr(prhs[7]);
	    x.Sdb = mxGetPr(prhs[8]);
	    x.Sindimage = mxGetPr(prhs[9]);
	    patch_pixels = x.pN;
	}

	x.rMat = new double[x.patch_w*x.patch_w];
	x.cMat = new double[x.patch_w*x.patch_w];

	x.patchS = new double[patch_pixels*3];
	if(x.patchS == NULL)
	    throw("Could not allocate x.patchS\n");

	// define working arrays
	std::list<std::pair<double,double> > * votes = new std::list<std::pair<double,double> >[x.TR*x.TC*3];

	// define output arrays
	mwSize sizeOut[3];
	sizeOut[0] = x.TR;
	sizeOut[1] = x.TC;
	sizeOut[2] = 3;

	plhs[0] = mxCreateNumericArray(3,sizeOut,mxDOUBLE_CLASS,mxREAL); //initializes to 0
	if(plhs[0] == NULL)
	    throw("Could not allocate output T\n");
	double * T = mxGetPr(plhs[0]);

	plhs[1] = mxCreateNumericArray(2,sizeOut,mxDOUBLE_CLASS,mxREAL); //initializes to 0
	if(plhs[1] == NULL)
	    throw("Could not allocate output Tweight\n");
	double * Tweight_out = mxGetPr(plhs[1]);

	if ( use_patchdb>0 )
	{
		if( transformation==1 )
			transformic_vote_weightedmedian_db<patch_rot_db>(x,nnfield,votes,Tweight_out);
		else if( transformation==2 )
			transformic_vote_weightedmedian_db<patch_sc_db>(x,nnfield,votes,Tweight_out);
		else if( transformation==3 )
			transformic_vote_weightedmedian_db<patch_sim_db>(x,nnfield,votes,Tweight_out);
		else if( transformation==4 )
			transformic_vote_weightedmedian_db<patch_fl_db>(x,nnfield,votes,Tweight_out);
		else if( transformation==5 )
			transformic_vote_weightedmedian_db<patch_il_db>(x,nnfield,votes,Tweight_out);
		else if( transformation==6 )
			transformic_vote_weightedmedian_db<patch_simb_db>(x,nnfield,votes,Tweight_out);
		else 
			transformic_vote_weightedmedian_db<patch_tr_db>(x,nnfield,votes,Tweight_out); // transformation 0
	}

	// for each pixel in T
	for(int k=0; k<3; k++)
	{
	    for(int c=0; c<x.TC; c++)
	    {
		for(int r=0; r<x.TR; r++)
		{
		    if( (Tweight_out[index2(r,c,x.TR)]>0) && (x.Tdomain[index2(r,c,x.TR)]>0.0) )
		    {
			votes[index3(r,c,k,x.TR,x.TC)].sort(compare_votes);

			std::list<std::pair<double,double> >::iterator it = votes[index3(r,c,k,x.TR,x.TC)].begin();
			double Tweight_curr = it->second;
			while( Tweight_curr < (Tweight_out[index2(r,c,x.TR)]/2.0) )
			{
			    it++;
			    Tweight_curr += it->second;
			}

			T[index3(r,c,k,x.TR,x.TC)] = it->first;
		    }
		}
	    }
	}

	delete [] votes;
	delete [] x.rMat;
	delete [] x.cMat;
	delete [] x.patchS;

	return;

    } 
    catch(std::bad_alloc&)
    {
	mexErrMsgTxt("Out of memory");
    }
    catch(const char* p)
    {
	std::string strOut(p);
	strOut += "\n";
	strOut += strUsage;
	mexErrMsgTxt(strOut.c_str());
    }
    catch(...)
    {
	mexErrMsgTxt("Unknown exception caught");
    }

}
