/* B = vgg_interp2(A, X, Y[, method[, oobv]])
 *
 *$Id: vgg_interp2.cxx,v 1.36 2008/03/30 18:49:28 ojw Exp $
 * Rewritten by ojw 20/9/06
 *
 * Edited and simplified by Alex Mansfield 6/12/10
 * for transformic
 *
 * License:
 *
 * Copyright (c) 2007-2008, University of Oxford
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *   * Neither the name of the University of Oxford nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <math.h>
#include <limits>
#ifdef _OPENMP
#include <omp.h>
#endif

// Define types
#ifdef _MSC_VER
typedef __int16 int16_t;
typedef unsigned __int8 uint8_t;
typedef unsigned __int16 uint16_t;
#else
#include <stdint.h>
#endif

template<class T, class U> static inline void interp_linear(U * const B, T const * const A, double const * const X, double const * const Y, int const num_points, int const w, int const h, int const col, U const oobv);

void vgg_interp2_c_db(double * const B, double const * const A, int const patch_w, int const num_points, double const * const X, double const * const Y)
{
    // B must be 3 x pN
	
    // Get the out of bounds value (oobv) and set the output class to the same class as the oobv.
  double oobv = (double)std::numeric_limits<double>::max();
	
    interp_linear(B, A, X, Y, num_points, patch_w, patch_w, 3, oobv);

    return;
}

// Function for correct rounding
// Add these to use numeric_limits class
using namespace std;
template<class U, class T> static inline U saturate_cast(T val)
{
    if (numeric_limits<U>::is_integer && !numeric_limits<T>::is_integer) {
	if (numeric_limits<U>::is_signed)
	    return val > 0 ? (val > (T)numeric_limits<U>::max() ? numeric_limits<U>::max() : static_cast<U>(val + 0.5)) : (val < (T)numeric_limits<U>::min() ? numeric_limits<U>::min() : static_cast<U>(val - 0.5));
	else
	    return val > 0 ? (val > (T)numeric_limits<U>::max() ? numeric_limits<U>::max() : static_cast<U>(val + 0.5)) : 0;
    }
    return static_cast<U>(val);
}

// Linear interpolation
template<class T, class U> static inline void interp_linear(U * const B, T const * const A, double const * const X, double const * const Y, int const num_points, int const w, int const h, int const col, U const oobv)
{
    double dw = (double)w;
    double dh = (double)h;

    // For each of the interpolation points
    int i, j, k, x, y;
    double u, v, out;
//#pragma omp parallel for if (num_points > 300) num_threads(omp_get_num_procs()) default(shared) private(i,j,k,u,v,x,y,out)
    for (i = 0; i < num_points; i++) {
	int it = 3*i;

	if (X[i] >= 1 && Y[i] >= 1) {
	    if (X[i] < dw) {
		if (Y[i] < dh) {
		    // Linearly interpolate
		    x = (int)X[i];
		    y = (int)Y[i];
		    u = X[i] - x;
		    v = Y[i] - y;
		    k = 3*(h * (x - 1) + y - 1);
		    for (j = it; j < (it+3); j++, k++) {
			out = A[k] + (A[k+(3*h)] - A[k]) * u;
			out += ((A[k+3] - out) + (A[k+(3*h)+3] - A[k+3]) * u) * v;
			B[j] = saturate_cast<U, double>(out);
		    }
		} else if (Y[i] == dh) {
		    // The Y coordinate is on the boundary
		    // Avoid reading outside the buffer to avoid crashes
		    // Linearly interpolate along X
		    x = (int)X[i];
		    u = X[i] - x;
		    k = 3*(h * x - 1);
		    for (j = it; j < (it+3); j++, k++)
			B[j] = saturate_cast<U, double>(A[k] + (A[k+(3*h)] - A[k]) * u);
		} else {
		    // Out of bounds
		    for (j = it; j < (it+3); j++)
			B[j] = oobv;
		}
	    } else if (X[i] == dw) {
		if (Y[i] < dh) {
		    // The X coordinate is on the boundary
		    // Avoid reading outside the buffer to avoid crashes
		    // Linearly interpolate along Y
		    y = (int)Y[i];
		    v = Y[i] - y;
		    k = 3*(h * (w - 1) + y - 1);
		    for (j = it; j < (it+3); j++, k++)
			B[j] = saturate_cast<U, double>(A[k] + (A[k+3] - A[k]) * v);
		} else if (Y[i] == dh) {
		    // The X and Y coordinates are on the boundary
		    // Avoid reading outside the buffer to avoid crashes
		    // Output the last value in the array
		    k = 3*(h * w - 1);
		    for (j = it; j < (it+3); j++, k++)
			B[j] = saturate_cast<U, double>(A[k]);
		} else {
		    // Out of bounds
		    for (j = it; j < (it+3); j++)
			B[j] = oobv;
		}
	    } else {
		// Out of bounds
		for (j = it; j < (it+3); j++)
		    B[j] = oobv;
	    }
	} else {
	    // Out of bounds
	    for (j = it; j < (it+3); j++)
		B[j] = oobv;
	}
    }
    return;
}
